#   Copyright 2024 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import logging

from collections.abc import Callable
from typing import Literal

import arviz as az
import numpy as np
import pymc as pm
import pytensor
import pytensor.tensor as pt
import xarray as xr

from arviz import dict_to_dataset
from better_optimize.constants import minimize_method
from numpy.typing import ArrayLike
from pymc import Model
from pymc.backends.arviz import coords_and_dims_for_inferencedata
from pymc.blocking import DictToArrayBijection
from pymc.model.transform.optimization import freeze_dims_and_data
from pymc.util import get_untransformed_name, is_transformed_name
from pytensor.graph import vectorize_graph
from pytensor.tensor import TensorVariable
from pytensor.tensor.optimize import minimize
from xarray import Dataset

from pymc_extras.inference.laplace_approx.find_map import (
    _compute_inverse_hessian,
    _make_initial_point,
    find_MAP,
)
from pymc_extras.inference.laplace_approx.scipy_interface import (
    GradientBackend,
    scipy_optimize_funcs_from_loss,
)

_log = logging.getLogger(__name__)


def get_conditional_gaussian_approximation(
    x: TensorVariable,
    Q: TensorVariable | ArrayLike,
    mu: TensorVariable | ArrayLike,
    args: list[TensorVariable] | None = None,
    model: pm.Model | None = None,
    method: minimize_method = "BFGS",
    use_jac: bool = True,
    use_hess: bool = False,
    optimizer_kwargs: dict | None = None,
) -> Callable:
    """
    Returns a function to estimate the a posteriori log probability of a latent Gaussian field x and its mode x0 using the Laplace approximation.

    That is:
    y | x, sigma ~ N(Ax, sigma^2 W)
    x | params ~ N(mu, Q(params)^-1)

    We seek to estimate log(p(x | y, params)):

    log(p(x | y, params)) = log(p(y | x, params)) + log(p(x | params)) + const

    Let f(x) = log(p(y | x, params)). From the definition of our model above, we have log(p(x | params)) = -0.5*(x - mu).T Q (x - mu) + 0.5*logdet(Q).

    This gives log(p(x | y, params)) = f(x) - 0.5*(x - mu).T Q (x - mu) + 0.5*logdet(Q). We will estimate this using the Laplace approximation by Taylor expanding f(x) about the mode.

    Thus:

    1. Maximize log(p(x | y, params)) = f(x) - 0.5*(x - mu).T Q (x - mu) wrt x (note that logdet(Q) does not depend on x) to find the mode x0.

    2. Substitute x0 into the Laplace approximation expanded about the mode: log(p(x | y, params)) ~= -0.5*x.T (-f''(x0) + Q) x + x.T (Q.mu + f'(x0) - f''(x0).x0) + 0.5*logdet(Q).

    Parameters
    ----------
    x: TensorVariable
        The parameter with which to maximize wrt (that is, find the mode in x). In INLA, this is the latent field x~N(mu,Q^-1).
    Q: TensorVariable | ArrayLike
        The precision matrix of the latent field x.
    mu: TensorVariable | ArrayLike
        The mean of the latent field x.
    args: list[TensorVariable]
        Args to supply to the compiled function. That is, (x0, logp) = f(x, *args). If set to None, assumes the model RVs are args.
    model: Model
        PyMC model to use.
    method: minimize_method
        Which minimization algorithm to use.
    use_jac: bool
        If true, the minimizer will compute the gradient of log(p(x | y, params)).
    use_hess: bool
        If true, the minimizer will compute the Hessian log(p(x | y, params)).
    optimizer_kwargs: dict
        Kwargs to pass to scipy.optimize.minimize.

    Returns
    -------
    f: Callable
        A function which accepts a value of x and args and returns [x0, log(p(x | y, params))], where x0 is the mode. x is currently both the point at which to evaluate logp and the initial guess for the minimizer.
    """
    model = pm.modelcontext(model)

    if args is None:
        args = model.continuous_value_vars + model.discrete_value_vars

    # f = log(p(y | x, params))
    f_x = model.logp()
    jac = pytensor.gradient.grad(f_x, x)
    hess = pytensor.gradient.jacobian(jac.flatten(), x)

    # log(p(x | y, params)) only including terms that depend on x for the minimization step (logdet(Q) ignored as it is a constant wrt x)
    log_x_posterior = f_x - 0.5 * (x - mu).T @ Q @ (x - mu)

    # Maximize log(p(x | y, params)) wrt x to find mode x0
    x0, _ = minimize(
        objective=-log_x_posterior,
        x=x,
        method=method,
        jac=use_jac,
        hess=use_hess,
        optimizer_kwargs=optimizer_kwargs,
    )

    # require f'(x0) and f''(x0) for Laplace approx
    jac = pytensor.graph.replace.graph_replace(jac, {x: x0})
    hess = pytensor.graph.replace.graph_replace(hess, {x: x0})

    # Full log(p(x | y, params)) using the Laplace approximation (up to a constant)
    _, logdetQ = pt.linalg.slogdet(Q)
    conditional_gaussian_approx = (
        -0.5 * x.T @ (-hess + Q) @ x + x.T @ (Q @ mu + jac - hess @ x0) + 0.5 * logdetQ
    )

    # Currently x is passed both as the query point for f(x, args) = logp(x | y, params) AND as an initial guess for x0. This may cause issues if the query point is
    # far from the mode x0 or in a neighbourhood which results in poor convergence.
    return pytensor.function(args, [x0, conditional_gaussian_approx])


def unpack_last_axis(packed_input, packed_shapes):
    if len(packed_shapes) == 1:
        # Single case currently fails in unpack
        return [pt.split_dims(packed_input, packed_shapes[0], axis=-1)]

    keep_axes = tuple(range(packed_input.ndim))[:-1]
    return pt.unpack(packed_input, keep_axes=keep_axes, packed_shapes=packed_shapes)


def draws_from_laplace_approx(
    *,
    mean,
    covariance=None,
    standard_deviation=None,
    draws: int,
    model: Model,
    vectorize_draws: bool = True,
    return_unconstrained: bool = True,
    random_seed=None,
    compile_kwargs: dict | None = None,
) -> tuple[Dataset, Dataset | None]:
    """
    Generate draws from the Laplace approximation of the posterior.

    Parameters
    ----------
    mean : np.ndarray
        The mean of the Laplace approximation (MAP estimate).
    covariance : np.ndarray, optional
        The covariance matrix of the Laplace approximation.
        Mutually exclusive with `standard_deviation`.
    standard_deviation : np.ndarray, optional
        The standard deviation of the Laplace approximation (diagonal approximation).
        Mutually exclusive with `covariance`.
    draws : int
        The number of draws.
    model : pm.Model
        The PyMC model.
    vectorize_draws : bool, default True
        Whether to vectorize the draws.
    return_unconstrained : bool, default True
        Whether to return the unconstrained draws in addition to the constrained ones.
    random_seed : int, optional
        Random seed for reproducibility.
    compile_kwargs: dict, optional
        Optional compile kwargs

    Returns
    -------
    tuple[Dataset, Dataset | None]
        A tuple containing the constrained draws (trace) and optionally the unconstrained draws.

    Raises
    ------
    ValueError
        If neither `covariance` nor `standard_deviation` is provided,
        or if both are provided.
    """
    # This function assumes that mean/covariance/standard_deviation are aligned with model.initial_point()
    if covariance is None and standard_deviation is None:
        raise ValueError("Must specify either covariance or standard_deviation")
    if covariance is not None and standard_deviation is not None:
        raise ValueError("Cannot specify both covariance and standard_deviation")
    if compile_kwargs is None:
        compile_kwargs = {}

    initial_point = model.initial_point()
    n = int(np.sum([np.prod(v.shape) for v in initial_point.values()]))
    assert mean.shape == (n,)
    if covariance is not None:
        assert covariance.shape == (n, n)
    elif standard_deviation is not None:
        assert standard_deviation.shape == (n,)

    vars_to_sample = [v for v in model.free_RVs + model.deterministics]
    var_names = [v.name for v in vars_to_sample]

    orig_constrained_vars = model.value_vars
    orig_outputs = model.replace_rvs_by_values(vars_to_sample)
    if return_unconstrained:
        orig_outputs.extend(model.value_vars)

    mu_pt = pt.vector("mu", shape=(n,), dtype=mean.dtype)
    size = (draws,) if vectorize_draws else ()
    if covariance is not None:
        sigma_pt = pt.matrix("cov", shape=(n, n), dtype=covariance.dtype)
        laplace_approximation = pm.MvNormal.dist(mu=mu_pt, cov=sigma_pt, size=size, method="svd")
    else:
        sigma_pt = pt.vector("sigma", shape=(n,), dtype=standard_deviation.dtype)
        laplace_approximation = pm.Normal.dist(mu=mu_pt, sigma=sigma_pt, size=(*size, n))

    constrained_vars = unpack_last_axis(
        laplace_approximation,
        [initial_point[v.name].shape for v in orig_constrained_vars],
    )
    outputs = vectorize_graph(
        orig_outputs, replace=dict(zip(orig_constrained_vars, constrained_vars))
    )

    fn = pm.pytensorf.compile(
        [mu_pt, sigma_pt],
        outputs,
        random_seed=random_seed,
        trust_input=True,
        **compile_kwargs,
    )
    sigma = covariance if covariance is not None else standard_deviation
    if vectorize_draws:
        output_buffers = fn(mean, sigma)
    else:
        # Take one draw to find the shape of the outputs
        output_buffers = []
        for out_draw in fn(mean, sigma):
            output_buffer = np.empty((draws, *out_draw.shape), dtype=out_draw.dtype)
            output_buffer[0] = out_draw
            output_buffers.append(output_buffer)
        # Fill one draws at a time
        for i in range(1, draws):
            for out_buffer, out_draw in zip(output_buffers, fn(mean, sigma)):
                out_buffer[i] = out_draw

    model_coords, model_dims = coords_and_dims_for_inferencedata(model)
    posterior = {
        var_name: out_buffer[None]
        for var_name, out_buffer in (
            zip(var_names, output_buffers, strict=not return_unconstrained)
        )
    }
    posterior_dataset = dict_to_dataset(posterior, coords=model_coords, dims=model_dims, library=pm)
    unconstrained_posterior_dataset = None

    if return_unconstrained:
        unconstrained_posterior = {
            var.name: out_buffer[None]
            for var, out_buffer in zip(
                model.value_vars, output_buffers[len(posterior) :], strict=True
            )
        }
        # Attempt to map constrained dims to unconstrained dims
        for var_name, var_draws in unconstrained_posterior.items():
            if not is_transformed_name(var_name):
                # constrained == unconstrained, dims already shared
                continue
            constrained_dims = model_dims.get(get_untransformed_name(var_name))
            if constrained_dims is None or (len(constrained_dims) != (var_draws.ndim - 2)):
                continue
            # Reuse dims from constrained variable if they match in length with unconstrained draws
            inferred_dims = []
            for i, (constrained_dim, unconstrained_dim_length) in enumerate(
                zip(constrained_dims, var_draws.shape[2:], strict=True)
            ):
                if model_coords.get(constrained_dim) is not None and (
                    len(model_coords[constrained_dim]) == unconstrained_dim_length
                ):
                    # Assume coordinates map. This could be fooled, by e.g., having a transform that reverses values
                    inferred_dims.append(constrained_dim)
                else:
                    # Size mismatch (e.g., Simplex), make no assumption about mapping
                    inferred_dims.append(f"{var_name}_dim_{i}")
            model_dims[var_name] = inferred_dims

        unconstrained_posterior_dataset = dict_to_dataset(
            unconstrained_posterior,
            coords=model_coords,
            dims=model_dims,
            library=pm,
        )

    return posterior_dataset, unconstrained_posterior_dataset


def fit_laplace(
    optimize_method: minimize_method | Literal["basinhopping"] = "BFGS",
    *,
    model: pm.Model | None = None,
    use_grad: bool | None = None,
    use_hessp: bool | None = None,
    use_hess: bool | None = None,
    initvals: dict | None = None,
    random_seed: int | np.random.Generator | None = None,
    jitter_rvs: list[pt.TensorVariable] | None = None,
    progressbar: bool = True,
    include_transformed: bool = True,
    freeze_model: bool = True,
    gradient_backend: GradientBackend = "pytensor",
    chains: None | int = None,
    draws: int = 500,
    vectorize_draws: bool = True,
    optimizer_kwargs: dict | None = None,
    compile_kwargs: dict | None = None,
) -> az.InferenceData:
    """
    Create a Laplace (quadratic) approximation for a posterior distribution.

    This function generates a Laplace approximation for a given posterior distribution using a specified
    number of draws. This is useful for obtaining a parametric approximation to the posterior distribution
    that can be used for further analysis.

    Parameters
    ----------
    model : pm.Model
        The PyMC model to be fit. If None, the current model context is used.
    optimize_method : str
        The optimization method to use. Valid choices are: Nelder-Mead, Powell, CG, BFGS, L-BFGS-B, TNC, SLSQP,
        trust-constr, dogleg, trust-ncg, trust-exact, trust-krylov, and basinhopping.

        See scipy.optimize.minimize documentation for details.
    use_grad : bool | None, optional
        Whether to use gradients in the optimization. Defaults to None, which determines this automatically based on
        the ``method``.
    use_hessp : bool | None, optional
        Whether to use Hessian-vector products in the optimization. Defaults to None, which determines this automatically based on
        the ``method``.
    use_hess : bool | None, optional
        Whether to use the Hessian matrix in the optimization. Defaults to None, which determines this automatically based on
        the ``method``.
    initvals : None | dict, optional
        Initial values for the model parameters, as str:ndarray key-value pairs. Paritial initialization is permitted.
         If None, the model's default initial values are used.
    random_seed : None | int | np.random.Generator, optional
        Seed for the random number generator or a numpy Generator for reproducibility
    jitter_rvs : list of TensorVariables, optional
        Variables whose initial values should be jittered. If None, all variables are jittered.
    progressbar : bool, optional
        Whether to display a progress bar during optimization. Defaults to True.
    include_transformed: bool, default True
        Whether to include transformed variables in the output. If True, transformed variables will be included in the
        output InferenceData object. If False, only the original variables will be included.
    freeze_model: bool, optional
        If True, freeze_dims_and_data will be called on the model before compiling the loss functions. This is
        sometimes necessary for JAX, and can sometimes improve performance by allowing constant folding. Defaults to
        True.
    gradient_backend: str, default "pytensor"
        The backend to use for gradient computations. Must be one of "pytensor" or "jax".
    draws: int, default: 500
        The number of samples to draw from the approximated posterior.
    optimizer_kwargs
        Additional keyword arguments to pass to the ``scipy.optimize`` function being used. Unless
        ``method = "basinhopping"``, ``scipy.optimize.minimize`` will be used. For ``basinhopping``,
        ``scipy.optimize.basinhopping`` will be used. See the documentation of these functions for details.
    vectorize_draws: bool, default True
        Whether to natively vectorize the random function or take one at a time in a python loop.
    compile_kwargs: dict, optional
        Additional keyword arguments to pass to pytensor.function.

    Returns
    -------
    :class:`~arviz.InferenceData`
        An InferenceData object containing the approximated posterior samples.

    Examples
    --------
    >>> from pymc_extras.inference import fit_laplace
    >>> import numpy as np
    >>> import pymc as pm
    >>> import arviz as az
    >>> y = np.array([2642, 3503, 4358] * 10)
    >>> with pm.Model() as m:
    >>>     logsigma = pm.Uniform("logsigma", 1, 100)
    >>>     mu = pm.Uniform("mu", -10000, 10000)
    >>>     yobs = pm.Normal("y", mu=mu, sigma=pm.math.exp(logsigma), observed=y)
    >>>     idata = fit_laplace()

    Notes
    -----
    This method of approximation may not be suitable for all types of posterior distributions,
    especially those with significant skewness or multimodality.

    See Also
    --------
    fit : Calling the inference function 'fit' like pmx.fit(method="laplace", model=m)
          will forward the call to 'fit_laplace'.

    """
    if chains is not None:
        raise ValueError(
            "chains argument has been deprecated. "
            "The behavior can be recreated by unstacking draws into multiple chains after fitting"
        )

    compile_kwargs = {} if compile_kwargs is None else compile_kwargs
    optimizer_kwargs = {} if optimizer_kwargs is None else optimizer_kwargs
    model = pm.modelcontext(model) if model is None else model

    if freeze_model:
        model = freeze_dims_and_data(model)

    idata = find_MAP(
        method=optimize_method,
        model=model,
        use_grad=use_grad,
        use_hessp=use_hessp,
        use_hess=use_hess,
        initvals=initvals,
        random_seed=random_seed,
        jitter_rvs=jitter_rvs,
        progressbar=progressbar,
        include_transformed=include_transformed,
        freeze_model=False,
        gradient_backend=gradient_backend,
        compile_kwargs=compile_kwargs,
        compute_hessian=True,
        **optimizer_kwargs,
    )

    if "covariance_matrix" not in idata.fit:
        # The user didn't use `use_hess` or `use_hessp` (or an optimization method that returns an inverse Hessian), so
        # we have to go back and compute the Hessian at the MAP point now.
        unpacked_variable_names = idata.fit["mean_vector"].coords["rows"].values.tolist()
        frozen_model = freeze_dims_and_data(model)
        initial_params = _make_initial_point(frozen_model, initvals, random_seed, jitter_rvs)

        _, f_hessp = scipy_optimize_funcs_from_loss(
            loss=-frozen_model.logp(jacobian=False),
            inputs=frozen_model.continuous_value_vars + frozen_model.discrete_value_vars,
            initial_point_dict=DictToArrayBijection.rmap(initial_params),
            use_grad=False,
            use_hess=False,
            use_hessp=True,
            gradient_backend=gradient_backend,
            compile_kwargs=compile_kwargs,
        )
        H_inv = _compute_inverse_hessian(
            optimizer_result=None,
            optimal_point=idata.fit.mean_vector.values,
            f_fused=None,
            f_hessp=f_hessp,
            use_hess=False,
            method=optimize_method,
        )

        idata.fit["covariance_matrix"] = xr.DataArray(
            H_inv,
            dims=("rows", "columns"),
            coords={"rows": unpacked_variable_names, "columns": unpacked_variable_names},
        )

    # We override the posterior/unconstrained_posterior from find_MAP
    idata.posterior, unconstrained_posterior = draws_from_laplace_approx(
        mean=idata.fit["mean_vector"].values,
        covariance=idata.fit["covariance_matrix"].values,
        draws=draws,
        return_unconstrained=include_transformed,
        model=model,
        vectorize_draws=vectorize_draws,
        random_seed=random_seed,
    )
    if include_transformed:
        idata.unconstrained_posterior = unconstrained_posterior
    return idata



"""
Tools and utilities shared by statespace tests.

There are no actual tests in this file -- the name is chosen to trigger automatic discovery of the sub-folders by
pytest.
"""

import numpy as np
import pandas as pd
import pytensor
import pytensor.tensor as pt
import statsmodels.api as sm

from numpy.testing import assert_allclose
from pymc import modelcontext

from pymc_extras.statespace.filters.kalman_smoother import KalmanSmoother
from pymc_extras.statespace.utils.constants import (
    MATRIX_NAMES,
    SHORT_NAME_TO_LONG,
)
from tests.statespace.statsmodel_local_level import LocalLinearTrend

floatX = pytensor.config.floatX


def load_nile_test_data():
    from importlib.metadata import version

    nile = pd.read_csv("tests/statespace/_data/nile.csv", dtype={"x": floatX})
    major, minor, rev = map(int, version("pandas").split("."))
    if major >= 3 or (major >= 2 and minor >= 2 and rev >= 0):
        freq_str = "YS-JAN"
    else:
        freq_str = "AS-JAN"
    nile.index = pd.date_range(start="1871-01-01", end="1970-01-01", freq=freq_str)
    nile.rename(columns={"x": "height"}, inplace=True)
    nile = (nile - nile.mean()) / nile.std()
    nile = nile.astype(floatX)

    return nile


def initialize_filter(kfilter, p=None, m=None, r=None, n=None):
    ksmoother = KalmanSmoother()
    data = pt.tensor(name="data", dtype=floatX, shape=(n, p))
    a0 = pt.tensor(name="x0", dtype=floatX, shape=(m,))
    P0 = pt.tensor(name="P0", dtype=floatX, shape=(m, m))
    c = pt.tensor(name="c", dtype=floatX, shape=(m,))
    d = pt.tensor(name="d", dtype=floatX, shape=(p,))
    Q = pt.tensor(name="Q", dtype=floatX, shape=(r, r))
    H = pt.tensor(name="H", dtype=floatX, shape=(p, p))
    T = pt.tensor(name="T", dtype=floatX, shape=(m, m))
    R = pt.tensor(name="R", dtype=floatX, shape=(m, r))
    Z = pt.tensor(name="Z", dtype=floatX, shape=(p, m))

    inputs = [data, a0, P0, c, d, T, Z, R, H, Q]

    (
        filtered_states,
        predicted_states,
        observed_states,
        filtered_covs,
        predicted_covs,
        observed_covs,
        ll_obs,
    ) = kfilter.build_graph(*inputs)

    smoothed_states, smoothed_covs = ksmoother.build_graph(T, R, Q, filtered_states, filtered_covs)

    outputs = [
        filtered_states,
        predicted_states,
        smoothed_states,
        filtered_covs,
        predicted_covs,
        smoothed_covs,
        ll_obs.sum(),
        ll_obs,
    ]

    return inputs, outputs


def add_missing_data(data, n_missing, rng):
    n = data.shape[0]
    missing_idx = rng.choice(n, n_missing, replace=False)
    data[missing_idx] = np.nan

    return data


def make_test_inputs(p, m, r, n, rng, missing_data=None, H_is_zero=False):
    data = np.arange(n * p, dtype=floatX).reshape(-1, p)
    if missing_data is not None:
        data = add_missing_data(data, missing_data, rng)

    a0 = np.zeros(m, dtype=floatX)
    P0 = np.eye(m, dtype=floatX)
    c = np.zeros(m, dtype=floatX)
    d = np.zeros(p, dtype=floatX)
    Q = np.eye(r, dtype=floatX)
    H = np.zeros((p, p), dtype=floatX) if H_is_zero else np.eye(p, dtype=floatX)
    T = np.eye(m, k=-1, dtype=floatX)
    T[0, :] = 1 / m
    R = np.eye(m, dtype=floatX)[:, :r]
    Z = np.eye(m, dtype=floatX)[:p, :]

    data, a0, P0, c, d, T, Z, R, H, Q = map(
        np.ascontiguousarray, [data, a0, P0, c, d, T, Z, R, H, Q]
    )

    return data, a0, P0, c, d, T, Z, R, H, Q


def get_expected_shape(name, p, m, r, n):
    if name == "log_likelihood":
        return ()
    elif name == "ll_obs":
        return (n,)
    filter_type, variable = name.split("_")
    if variable == "states":
        return n, m
    if variable == "covs":
        return n, m, m


def get_sm_state_from_output_name(res, name):
    if name == "log_likelihood":
        return res.llf
    elif name == "ll_obs":
        return res.llf_obs

    filter_type, variable = name.split("_")
    sm_states = getattr(res, "states")

    if variable == "states":
        return getattr(sm_states, filter_type)
    if variable == "covs":
        m = res.filter_results.k_states
        # remove the "s" from "covs"
        return getattr(sm_states, name[:-1]).reshape(-1, m, m)


def nile_test_test_helper(rng, n_missing=0):
    a0 = np.zeros(2, dtype=floatX)
    P0 = np.eye(2, dtype=floatX) * 1e6
    c = np.zeros(2, dtype=floatX)
    d = np.zeros(1, dtype=floatX)
    Q = np.eye(2, dtype=floatX) * np.array([0.5, 0.01], dtype=floatX)
    H = np.eye(1, dtype=floatX) * 0.8
    T = np.array([[1.0, 1.0], [0.0, 1.0]], dtype=floatX)
    R = np.eye(2, dtype=floatX)
    Z = np.array([[1.0, 0.0]], dtype=floatX)

    # In pandas > 3.0, the array stored in .values is a read-only pointer, so we need to copy it
    data = load_nile_test_data().values.copy()
    if n_missing > 0:
        data = add_missing_data(data, n_missing, rng)

    sm_model = LocalLinearTrend(
        endog=data,
        initialization="known",
        initial_state_cov=P0,
        initial_state=a0.ravel(),
    )

    res = sm_model.fit_constrained(
        constraints={
            "sigma2.measurement": 0.8,
            "sigma2.level": 0.5,
            "sigma2.trend": 0.01,
        }
    )

    inputs = [data, a0, P0, c, d, T, Z, R, H, Q]

    return res, inputs


def fast_eval(var):
    return pytensor.function([], var, mode="FAST_COMPILE")()


def delete_rvs_from_model(rv_names: list[str]) -> None:
    """Remove all model mappings referring to rv

    This can be used to "delete" an RV from a model
    """
    mod = modelcontext(None)
    all_rvs = mod.basic_RVs + mod.deterministics
    all_rv_names = [x.name for x in all_rvs]

    for name in rv_names:
        assert name in all_rv_names, f"{name} is not part of the Model: {all_rv_names}"

        rv_idx = all_rv_names.index(name)
        rv = all_rvs[rv_idx]

        mod.named_vars.pop(name)
        if name in mod.named_vars_to_dims:
            mod.named_vars_to_dims.pop(name)

        if rv in mod.deterministics:
            mod.deterministics.remove(rv)
            continue

        value = mod.rvs_to_values.pop(rv)
        mod.values_to_rvs.pop(value)
        mod.rvs_to_transforms.pop(rv)
        if rv in mod.free_RVs:
            mod.free_RVs.remove(rv)
            mod.rvs_to_initial_values.pop(rv)
        else:
            mod.observed_RVs.remove(rv)


def unpack_statespace(ssm):
    return [ssm[SHORT_NAME_TO_LONG[x]] for x in MATRIX_NAMES]


def unpack_symbolic_matrices_with_params(mod, param_dict, data_dict=None, mode="FAST_COMPILE"):
    inputs = list(mod._name_to_variable.values())
    if data_dict is not None:
        inputs += list(mod._name_to_data.values())
    else:
        data_dict = {}

    f_matrices = pytensor.function(
        inputs,
        unpack_statespace(mod.ssm),
        on_unused_input="raise",
        mode=mode,
    )

    x0, P0, c, d, T, Z, R, H, Q = f_matrices(**param_dict, **data_dict)

    return x0, P0, c, d, T, Z, R, H, Q


def simulate_from_numpy_model(mod, rng, param_dict, data_dict=None, steps=100):
    """
    Helper function to visualize the components outside of a PyMC model context
    """
    x0, P0, c, d, T, Z, R, H, Q = unpack_symbolic_matrices_with_params(mod, param_dict, data_dict)
    k_endog = mod.k_endog
    k_states = mod.k_states
    k_posdef = mod.k_posdef

    x = np.zeros((steps, k_states))
    y = np.zeros((steps, k_endog))

    x[0] = x0
    y[0] = (Z @ x0).squeeze() if Z.ndim == 2 else (Z[0] @ x0).squeeze()

    if not np.allclose(H, 0):
        y[0] += rng.multivariate_normal(mean=np.zeros(1), cov=H).squeeze()

    for t in range(1, steps):
        if k_posdef > 0:
            shock = rng.multivariate_normal(mean=np.zeros(k_posdef), cov=Q)
            innov = R @ shock
        else:
            innov = 0

        if not np.allclose(H, 0):
            error = rng.multivariate_normal(mean=np.zeros(1), cov=H)
        else:
            error = 0

        x[t] = c + T @ x[t - 1] + innov
        if Z.ndim == 2:
            y[t] = (d + Z @ x[t] + error).squeeze()
        else:
            y[t] = (d + Z[t] @ x[t] + error).squeeze()

    return x, y.squeeze()


def assert_pattern_repeats(y, T, atol, rtol):
    val = np.diff(y.reshape(-1, T), axis=0)
    if floatX.endswith("64"):
        # Round this before going into the test, otherwise it behaves poorly (atol = inf)
        n_digits = len(str(1 / atol))
        val = np.round(val, n_digits)

    assert_allclose(
        val,
        0,
        err_msg="seasonal pattern does not repeat",
        atol=atol,
        rtol=rtol,
    )


def make_stationary_params(data, p, d, q, P, D, Q, S):
    sm_sarimax = sm.tsa.SARIMAX(data, order=(p, d, q), seasonal_order=(P, D, Q, S))
    res = sm_sarimax.fit(disp=False)

    param_dict = dict(ar_params=[], ma_params=[], seasonal_ar_params=[], seasonal_ma_params=[])

    for name, param in zip(res.param_names, res.params):
        if name.startswith("ar.S"):
            param_dict["seasonal_ar_params"].append(param)
        elif name.startswith("ma.S"):
            param_dict["seasonal_ma_params"].append(param)
        elif name.startswith("ar."):
            param_dict["ar_params"].append(param)
        elif name.startswith("ma."):
            param_dict["ma_params"].append(param)
        else:
            param_dict["sigma_state"] = param

    param_dict = {
        k: np.array(v, dtype=floatX)
        for k, v in param_dict.items()
        if isinstance(v, float) or len(v) > 0
    }
    return param_dict



from copy import deepcopy

import numpy as np
import pymc as pm
import pytensor.tensor as pt
import pytest
import xarray as xr

from graphviz.graphs import Digraph
from preliz.distributions import distributions as preliz_distributions
from pydantic import ValidationError
from pymc.model_graph import fast_eval
from xarray import DataArray

import pymc_extras.prior as pr

from pymc_extras.deserialize import (
    DESERIALIZERS,
    deserialize,
    register_deserialization,
)
from pymc_extras.prior import (
    Censored,
    MuAlreadyExistsError,
    Prior,
    Scaled,
    UnknownTransformError,
    UnsupportedDistributionError,
    UnsupportedParameterizationError,
    UnsupportedShapeError,
    VariableFactory,
    handle_dims,
    register_tensor_transform,
    sample_prior,
)
from pymc_extras.utils.model_equivalence import equivalent_models


@pytest.mark.parametrize(
    "x, dims, desired_dims, expected_fn",
    [
        (np.arange(3), "channel", "channel", lambda x: x),
        (np.arange(3), "channel", ("geo", "channel"), lambda x: x),
        (np.arange(3), "channel", ("channel", "geo"), lambda x: x[:, None]),
        (np.arange(3), "channel", ("x", "y", "channel", "geo"), lambda x: x[:, None]),
        (
            np.arange(3 * 2).reshape(3, 2),
            ("channel", "geo"),
            ("geo", "x", "y", "channel"),
            lambda x: x.T[:, None, None, :],
        ),
        (
            np.arange(4 * 2 * 3).reshape(4, 2, 3),
            ("channel", "geo", "store"),
            ("geo", "x", "store", "channel"),
            lambda x: x.swapaxes(0, 2).swapaxes(0, 1)[:, None, :, :],
        ),
    ],
    ids=[
        "same_dims",
        "different_dims",
        "dim_padding",
        "just_enough_dims",
        "transpose_and_padding",
        "swaps_and_padding",
    ],
)
def test_handle_dims(x, dims, desired_dims, expected_fn) -> None:
    result = handle_dims(x, dims, desired_dims)
    if isinstance(result, pt.TensorVariable):
        result = fast_eval(result)

    np.testing.assert_array_equal(result, expected_fn(x))


@pytest.mark.parametrize(
    "x, dims, desired_dims",
    [
        (np.ones(3), "channel", "something_else"),
        (np.ones((3, 2)), ("a", "b"), ("a", "B")),
    ],
    ids=["no_incommon", "some_incommon"],
)
def test_handle_dims_with_impossible_dims(x, dims, desired_dims) -> None:
    match = " are not a subset of the desired dims "
    with pytest.raises(UnsupportedShapeError, match=match):
        handle_dims(x, dims, desired_dims)


def test_missing_transform() -> None:
    match = r"Function 'foo_bar' not present in pytensor.tensor or pymc.math"
    with pytest.raises(UnknownTransformError, match=match):
        Prior("Normal", transform="foo_bar")


def test_getattr() -> None:
    assert pr.Normal() == Prior("Normal")


def test_import_directly() -> None:
    try:
        from pymc_extras.prior import Normal
    except Exception as e:
        pytest.fail(f"Unexpected exception: {e}")

    assert Normal() == Prior("Normal")


def test_import_incorrect_directly() -> None:
    match = "PyMC doesn't have a distribution of name 'SomeIncorrectDistribution'"
    with pytest.raises(UnsupportedDistributionError, match=match):
        from pymc_extras.prior import SomeIncorrectDistribution  # noqa: F401


def test_get_item() -> None:
    var = Prior("Normal", mu=0, sigma=1)

    assert var["mu"] == 0
    assert var["sigma"] == 1


def test_noncentered_needs_params() -> None:
    with pytest.raises(ValueError):
        Prior(
            "Normal",
            centered=False,
        )


def test_different_than_pymc_params() -> None:
    with pytest.raises(ValueError):
        Prior("Normal", mu=0, b=1)


def test_non_unique_dims() -> None:
    with pytest.raises(ValueError):
        Prior("Normal", mu=0, sigma=1, dims=("channel", "channel"))


def test_doesnt_check_validity_parameterization() -> None:
    try:
        Prior("Normal", mu=0, sigma=1, tau=1)
    except Exception as e:
        pytest.fail(f"Unexpected exception: {e}")


def test_doesnt_check_validity_values() -> None:
    try:
        Prior("Normal", mu=0, sigma=-1)
    except Exception as e:
        pytest.fail(f"Unexpected exception: {e}")


def test_preliz() -> None:
    var = Prior("Normal", mu=0, sigma=1)
    dist = var.preliz
    assert isinstance(dist, preliz_distributions.Distribution)


@pytest.mark.parametrize(
    "var, expected",
    [
        (Prior("Normal", mu=0, sigma=1), 'Prior("Normal", mu=0, sigma=1)'),
        (
            Prior("Normal", mu=Prior("Normal"), sigma=Prior("HalfNormal")),
            'Prior("Normal", mu=Prior("Normal"), sigma=Prior("HalfNormal"))',
        ),
        (Prior("Normal", dims="channel"), 'Prior("Normal", dims="channel")'),
        (
            Prior("Normal", mu=0, sigma=1, transform="sigmoid"),
            'Prior("Normal", mu=0, sigma=1, transform="sigmoid")',
        ),
    ],
)
def test_str(var, expected) -> None:
    assert str(var) == expected


@pytest.mark.parametrize(
    "var",
    [
        Prior("Normal", mu=0, sigma=1),
        Prior("Normal", mu=Prior("Normal"), sigma=Prior("HalfNormal"), dims="channel"),
        Prior("Normal", dims=("geo", "channel")),
    ],
)
def test_repr(var) -> None:
    assert eval(repr(var)) == var


def test_invalid_distribution() -> None:
    with pytest.raises(UnsupportedDistributionError):
        Prior("Invalid")


def test_broadcast_doesnt_work():
    with pytest.raises(UnsupportedShapeError):
        Prior(
            "Normal",
            mu=0,
            sigma=Prior("HalfNormal", sigma=1, dims="x"),
            dims="y",
        )


def test_dim_workaround_flaw() -> None:
    distribution = Prior(
        "Normal",
        mu=Prior("Normal"),
        sigma=Prior("HalfNormal"),
        dims="y",
    )

    try:
        distribution["mu"].dims = "x"
    except Exception as e:
        pytest.fail(f"Unexpected exception: {e}")

    with pytest.raises(UnsupportedShapeError):
        distribution._param_dims_work()


def test_noncentered_error() -> None:
    with pytest.raises(UnsupportedParameterizationError):
        Prior(
            "Gamma",
            mu=0,
            sigma=1,
            dims="x",
            centered=False,
        )


def test_create_variable_multiple_times() -> None:
    mu = Prior(
        "Normal",
        mu=Prior("Normal"),
        sigma=Prior("HalfNormal"),
        dims="channel",
        centered=False,
    )

    coords = {
        "channel": ["a", "b", "c"],
    }
    with pm.Model(coords=coords) as model:
        mu.create_variable("mu")
        mu.create_variable("mu_2")

    suffixes = [
        "",
        "_offset",
        "_mu",
        "_sigma",
    ]
    dims = [(3,), (3,), (), ()]

    for prefix in ["mu", "mu_2"]:
        for suffix, dim in zip(suffixes, dims, strict=False):
            assert fast_eval(model[f"{prefix}{suffix}"]).shape == dim


def test_create_variable() -> None:
    large_var = Prior(
        "Normal",
        mu=Prior(
            "Normal",
            mu=Prior("Normal", mu=1),
            sigma=Prior("HalfNormal"),
            dims="channel",
            centered=False,
        ),
        sigma=Prior("HalfNormal", sigma=Prior("HalfNormal"), dims="geo"),
        dims=("geo", "channel"),
    )

    coords = {
        "channel": ["a", "b", "c"],
        "geo": ["x", "y"],
    }
    with pm.Model(coords=coords) as model:
        large_var.create_variable("var")

    var_names = [
        "var",
        "var_mu",
        "var_sigma",
        "var_mu_offset",
        "var_mu_mu",
        "var_mu_sigma",
        "var_sigma_sigma",
    ]
    assert set(var.name for var in model.unobserved_RVs) == set(var_names)
    dims = [
        (2, 3),
        (3,),
        (2,),
        (3,),
        (),
        (),
        (),
    ]
    for var_name, dim in zip(var_names, dims, strict=False):
        assert fast_eval(model[var_name]).shape == dim


def test_transform() -> None:
    var = Prior("Normal", mu=0, sigma=1, transform="sigmoid")

    with pm.Model() as model:
        var.create_variable("var")

    var_names = [
        "var",
        "var_raw",
    ]
    dims = [
        (),
        (),
    ]
    for var_name, dim in zip(var_names, dims, strict=False):
        assert fast_eval(model[var_name]).shape == dim


def test_to_dict() -> None:
    large_var = Prior(
        "Normal",
        mu=Prior(
            "Normal",
            mu=Prior("Normal", mu=1),
            sigma=Prior("HalfNormal"),
            dims="channel",
            centered=False,
        ),
        sigma=Prior("HalfNormal", sigma=Prior("HalfNormal"), dims="geo"),
        dims=("geo", "channel"),
    )

    data = large_var.to_dict()
    assert data == {
        "dist": "Normal",
        "kwargs": {
            "mu": {
                "dist": "Normal",
                "kwargs": {
                    "mu": {
                        "dist": "Normal",
                        "kwargs": {
                            "mu": 1,
                        },
                    },
                    "sigma": {
                        "dist": "HalfNormal",
                    },
                },
                "centered": False,
                "dims": ("channel",),
            },
            "sigma": {
                "dist": "HalfNormal",
                "kwargs": {
                    "sigma": {
                        "dist": "HalfNormal",
                    },
                },
                "dims": ("geo",),
            },
        },
        "dims": ("geo", "channel"),
    }

    assert Prior.from_dict(data) == large_var


def test_to_dict_numpy() -> None:
    var = Prior("Normal", mu=np.array([0, 10, 20]), dims="channel")
    assert var.to_dict() == {
        "dist": "Normal",
        "kwargs": {
            "mu": [0, 10, 20],
        },
        "dims": ("channel",),
    }


@pytest.mark.parametrize("dims", (None, (), ("city",)))
def test_dict_round_trip_dims(dims) -> None:
    prior = Prior("Normal", dims=dims)
    d = prior.to_dict()

    expected_d = {"dist": "Normal"}
    if dims is not None:
        expected_d["dims"] = dims
    assert d == expected_d

    prior_again = Prior.from_dict(d)
    assert prior_again.dims == dims
    assert prior_again == prior
    assert prior_again.to_dict() == expected_d


def test_constrain_with_transform_error() -> None:
    var = Prior("Normal", transform="sigmoid")

    with pytest.raises(ValueError):
        var.constrain(lower=0, upper=1)


def test_constrain() -> None:
    var = Prior("Normal")

    new_var = var.constrain(lower=0, upper=1, mass=0.9545)
    np.testing.assert_allclose(new_var.parameters["mu"], 0.5, rtol=1e-4)
    np.testing.assert_allclose(new_var.parameters["sigma"], 0.25, rtol=1e-4)


def test_dims_change() -> None:
    var = Prior("Normal", mu=0, sigma=1)
    var.dims = "channel"

    assert var.dims == ("channel",)


def test_dims_change_error() -> None:
    mu = Prior("Normal", dims="channel")
    var = Prior("Normal", mu=mu, dims="channel")

    with pytest.raises(UnsupportedShapeError):
        var.dims = "geo"


def test_deepcopy() -> None:
    priors = {
        "alpha": Prior("Beta", alpha=1, beta=1),
        "gamma": Prior("Normal", mu=0, sigma=1),
    }

    new_priors = deepcopy(priors)
    priors["alpha"].dims = "channel"

    assert new_priors["alpha"].dims is None


def test_backwards_compat() -> None:
    """Make sure functionality is compatible with use in PyMC-marketing, where Prior objects originated from."""
    mmm_default_model_config = {
        "intercept": {"dist": "Normal", "kwargs": {"mu": 0, "sigma": 2}},
        "likelihood": {
            "dist": "Normal",
            "kwargs": {
                "sigma": {"dist": "HalfNormal", "kwargs": {"sigma": 2}},
            },
        },
        "gamma_control": {
            "dist": "Normal",
            "kwargs": {"mu": 0, "sigma": 2},
            "dims": "control",
        },
        "gamma_fourier": {
            "dist": "Laplace",
            "kwargs": {"mu": 0, "b": 1},
            "dims": "fourier_mode",
        },
    }

    result = {param: Prior.from_dict(value) for param, value in mmm_default_model_config.items()}
    assert result == {
        "intercept": Prior("Normal", mu=0, sigma=2),
        "likelihood": Prior("Normal", sigma=Prior("HalfNormal", sigma=2)),
        "gamma_control": Prior("Normal", mu=0, sigma=2, dims="control"),
        "gamma_fourier": Prior("Laplace", mu=0, b=1, dims="fourier_mode"),
    }


def test_sample_prior() -> None:
    var = Prior(
        "Normal",
        mu=Prior("Normal"),
        sigma=Prior("HalfNormal"),
        dims="channel",
        transform="sigmoid",
    )

    coords = {"channel": ["A", "B", "C"]}
    prior = var.sample_prior(coords=coords, draws=25)

    assert isinstance(prior, xr.Dataset)
    assert prior.sizes == {"chain": 1, "draw": 25, "channel": 3}


def test_sample_prior_missing_coords() -> None:
    dist = Prior("Normal", dims="channel")

    with pytest.raises(KeyError, match="Coords"):
        dist.sample_prior()


def test_to_graph() -> None:
    hierarchical_distribution = Prior(
        "Normal",
        mu=Prior("Normal"),
        sigma=Prior("HalfNormal"),
        dims="channel",
    )

    G = hierarchical_distribution.to_graph()
    assert isinstance(G, Digraph)


def test_from_dict_list() -> None:
    data = {
        "dist": "Normal",
        "kwargs": {
            "mu": [0, 1, 2],
            "sigma": 1,
        },
        "dims": "channel",
    }

    var = Prior.from_dict(data)
    assert var.dims == ("channel",)
    assert isinstance(var["mu"], np.ndarray)
    np.testing.assert_array_equal(var["mu"], [0, 1, 2])


def test_from_dict_list_dims() -> None:
    data = {
        "dist": "Normal",
        "kwargs": {
            "mu": 0,
            "sigma": 1,
        },
        "dims": ["channel", "geo"],
    }

    var = Prior.from_dict(data)
    assert var.dims == ("channel", "geo")


def test_to_dict_transform() -> None:
    dist = Prior("Normal", transform="sigmoid")

    data = dist.to_dict()
    assert data == {
        "dist": "Normal",
        "transform": "sigmoid",
    }


def test_equality_non_prior() -> None:
    dist = Prior("Normal")

    assert dist != 1


def test_deepcopy_memo() -> None:
    memo = {}
    dist = Prior("Normal")
    memo[id(dist)] = dist
    deepcopy(dist, memo)
    assert len(memo) == 1
    deepcopy(dist, memo)

    assert len(memo) == 1


def test_create_likelihood_variable() -> None:
    distribution = Prior("Normal", sigma=Prior("HalfNormal"))

    with pm.Model() as model:
        mu = pm.Normal("mu")

        data = distribution.create_likelihood_variable("data", mu=mu, observed=10)

    assert model.observed_RVs == [data]
    assert "data_sigma" in model


def test_create_likelihood_variable_already_has_mu() -> None:
    distribution = Prior("Normal", mu=Prior("Normal"), sigma=Prior("HalfNormal"))

    with pm.Model():
        mu = pm.Normal("mu")

        with pytest.raises(MuAlreadyExistsError):
            distribution.create_likelihood_variable("data", mu=mu, observed=10)


def test_create_likelihood_non_mu_parameterized_distribution() -> None:
    distribution = Prior("Cauchy")

    with pm.Model():
        mu = pm.Normal("mu")
        with pytest.raises(UnsupportedDistributionError):
            distribution.create_likelihood_variable("data", mu=mu, observed=10)


def test_non_centered_student_t() -> None:
    try:
        Prior(
            "StudentT",
            mu=Prior("Normal"),
            sigma=Prior("HalfNormal"),
            nu=Prior("HalfNormal"),
            dims="channel",
            centered=False,
        )
    except Exception as e:
        pytest.fail(f"Unexpected exception: {e}")


def test_cant_reset_distribution() -> None:
    dist = Prior("Normal")
    with pytest.raises(AttributeError, match="Can't change the distribution"):
        dist.distribution = "Cauchy"


def test_nonstring_distribution() -> None:
    with pytest.raises(ValidationError, match=".*Input should be a valid string.*"):
        Prior(pm.Normal)


def test_change_the_transform() -> None:
    dist = Prior("Normal")
    dist.transform = "logit"
    assert dist.transform == "logit"


def test_nonstring_transform() -> None:
    with pytest.raises(ValidationError, match=".*Input should be a valid string.*"):
        Prior("Normal", transform=pm.math.log)


def test_checks_param_value_types() -> None:
    with pytest.raises(ValueError, match="Parameters must be one of the following types"):
        Prior("Normal", mu="str", sigma="str")


def test_check_equality_with_numpy() -> None:
    dist = Prior("Normal", mu=np.array([1, 2, 3]), sigma=1)
    assert dist == dist.deepcopy()


def clear_custom_transforms() -> None:
    global CUSTOM_TRANSFORMS
    CUSTOM_TRANSFORMS = {}


def test_custom_transform() -> None:
    new_transform_name = "foo_bar"
    with pytest.raises(UnknownTransformError):
        Prior("Normal", transform=new_transform_name)

    register_tensor_transform(new_transform_name, lambda x: x**2)

    dist = Prior("Normal", transform=new_transform_name)
    prior = dist.sample_prior(draws=10)
    df_prior = prior.to_dataframe()

    np.testing.assert_array_equal(
        df_prior.variable.to_numpy(),
        df_prior.variable_raw.to_numpy() ** 2,
    )


def test_custom_transform_comes_first() -> None:
    # function in pytensor.tensor
    register_tensor_transform("square", lambda x: 2 * x)

    dist = Prior("Normal", transform="square")
    prior = dist.sample_prior(draws=10)
    df_prior = prior.to_dataframe()

    np.testing.assert_array_equal(
        df_prior.variable.to_numpy(),
        2 * df_prior.variable_raw.to_numpy(),
    )

    clear_custom_transforms()


def test_serialize_with_pytensor() -> None:
    sigma = pt.arange(1, 4)
    dist = Prior("Normal", mu=0, sigma=sigma)

    assert dist.to_dict() == {
        "dist": "Normal",
        "kwargs": {
            "mu": 0,
            "sigma": [1, 2, 3],
        },
    }


def test_zsn_non_centered() -> None:
    try:
        Prior("ZeroSumNormal", sigma=1, centered=False)
    except Exception as e:
        pytest.fail(f"Unexpected exception: {e}")


class TestCustomClass:
    def test_sample_prior_arbitrary_no_name(self) -> None:
        class ArbitraryWithoutName:
            def __init__(self, dims: str | tuple[str, ...]) -> None:
                self.dims = dims

            def create_variable(self, name: str, xdist: bool = False):
                if xdist:
                    raise NotImplementedError

                with pm.Model(name=name):
                    location = pm.Normal("location", dims=self.dims)
                    scale = pm.HalfNormal("scale", dims=self.dims)

                    return pm.Normal("standard_normal") * scale + location

        var = ArbitraryWithoutName(dims="channel")

        prior = sample_prior(var, coords={"channel": ["A", "B", "C"]}, draws=25)

        assert isinstance(prior, xr.Dataset)
        assert "variable" not in prior

        prior_with = sample_prior(
            var,
            coords={"channel": ["A", "B", "C"]},
            draws=25,
            wrap=True,
        )

        assert isinstance(prior_with, xr.Dataset)
        assert "variable" in prior_with

    def test_arbitrary_class(self) -> None:
        class Arbitrary:
            def __init__(self, dims: str | tuple[str, ...]) -> None:
                self.dims = dims

            def create_variable(self, name: str):
                return pm.Normal(name, dims=self.dims)

        prior = Prior(
            "Normal",
            mu=Arbitrary(dims=("channel",)),
            sigma=1,
            dims=("channel", "geo"),
        )

        coords = {
            "channel": ["C1", "C2", "C3"],
            "geo": ["G1", "G2"],
        }
        with pm.Model(coords=coords) as model:
            prior.create_variable("var")

        assert "var_mu" in model
        var_mu = model["var_mu"]
        assert fast_eval(var_mu).shape == (len(coords["channel"]),)

    def test_arbitrary_class_sample_prior(self) -> None:
        class Arbitrary:
            def __init__(self, dims: str | tuple[str, ...]) -> None:
                self.dims = dims

            def create_variable(self, name: str):
                return pm.Normal(name, dims=self.dims)

        var = Arbitrary(dims="channel")
        prior = sample_prior(var, coords={"channel": ["A", "B", "C"]}, draws=25)
        assert isinstance(prior, xr.Dataset)

    def test_arbitrary_serialization(self) -> None:
        class ArbitrarySerializable:
            def __init__(self, dims: str | tuple[str, ...]) -> None:
                self.dims = dims

            def create_variable(self, name: str):
                return pm.Normal(name, dims=self.dims)

            def to_dict(self):
                return {"dims": self.dims}

        arbitrary_serialized_data = {"dims": ("channel",)}

        register_deserialization(
            lambda data: isinstance(data, dict) and data.keys() == {"dims"},
            lambda data: ArbitrarySerializable(**data),
        )

        dist = Prior(
            "Normal",
            mu=ArbitrarySerializable(dims=("channel",)),
            sigma=1,
            dims=("channel", "geo"),
        )

        data = {
            "dist": "Normal",
            "kwargs": {
                "mu": arbitrary_serialized_data,
                "sigma": 1,
            },
            "dims": ("channel", "geo"),
        }

        assert dist.to_dict() == data

        dist_again = deserialize(data)
        assert isinstance(dist_again["mu"], ArbitrarySerializable)
        assert dist_again["mu"].dims == ("channel",)

        DESERIALIZERS.pop()


class TestScaled:
    def test_scaled_initializes_correctly(self) -> None:
        """Test that the Scaled class initializes correctly."""
        normal = Prior("Normal", mu=0, sigma=1)
        scaled = Scaled(normal, factor=2.0)

        assert scaled.dist == normal
        assert scaled.factor == 2.0

    def test_scaled_dims_property(self) -> None:
        """Test that the dims property returns the dimensions of the underlying distribution."""
        normal = Prior("Normal", mu=0, sigma=1, dims="channel")
        scaled = Scaled(normal, factor=2.0)

        assert scaled.dims == ("channel",)

        # Test with multiple dimensions
        normal.dims = ("channel", "geo")
        assert scaled.dims == ("channel", "geo")

    def test_scaled_create_variable(self) -> None:
        """Test that the create_variable method properly scales the variable."""
        normal = Prior("Normal", mu=0, sigma=1)
        scaled = Scaled(normal, factor=2.0)

        with pm.Model() as model:
            scaled_var = scaled.create_variable("scaled_var")

        # Check that both the scaled and unscaled variables exist
        assert "scaled_var" in model
        assert "scaled_var_unscaled" in model

        # The deterministic node should be the scaled variable
        assert model["scaled_var"] == scaled_var

    def test_scaled_creates_correct_dimensions(self) -> None:
        """Test that the scaled variable has the correct dimensions."""
        normal = Prior("Normal", dims="channel")
        scaled = Scaled(normal, factor=2.0)

        coords = {"channel": ["A", "B", "C"]}
        with pm.Model(coords=coords):
            scaled_var = scaled.create_variable("scaled_var")

        # Check that the scaled variable has the correct dimensions
        assert fast_eval(scaled_var).shape == (3,)

    def test_scaled_applies_factor(self) -> None:
        """Test that the scaling factor is correctly applied."""
        normal = Prior("Normal", mu=0, sigma=1)
        factor = 3.5
        scaled = Scaled(normal, factor=factor)

        # Sample from prior to verify scaling
        prior = sample_prior(scaled, draws=10, name="scaled_var")
        df_prior = prior.to_dataframe()

        # Check that scaled values are original values times the factor
        unscaled_values = df_prior["scaled_var_unscaled"].to_numpy()
        scaled_values = df_prior["scaled_var"].to_numpy()
        np.testing.assert_allclose(scaled_values, unscaled_values * factor)

    def test_scaled_with_tensor_factor(self) -> None:
        """Test that the Scaled class works with a tensor factor."""
        normal = Prior("Normal", mu=0, sigma=1)
        factor = pt.as_tensor_variable(2.5)
        scaled = Scaled(normal, factor=factor)

        # Sample from prior to verify tensor scaling
        prior = sample_prior(scaled, draws=10, name="scaled_var")
        df_prior = prior.to_dataframe()

        # Check that scaled values are original values times the factor
        unscaled_values = df_prior["scaled_var_unscaled"].to_numpy()
        scaled_values = df_prior["scaled_var"].to_numpy()
        np.testing.assert_allclose(scaled_values, unscaled_values * 2.5)

    def test_scaled_with_hierarchical_prior(self) -> None:
        """Test that the Scaled class works with hierarchical priors."""
        normal = Prior("Normal", mu=Prior("Normal"), sigma=Prior("HalfNormal"), dims="channel")
        scaled = Scaled(normal, factor=2.0)

        coords = {"channel": ["A", "B", "C"]}
        with pm.Model(coords=coords) as model:
            scaled.create_variable("scaled_var")

        # Check that all necessary variables were created
        assert "scaled_var" in model
        assert "scaled_var_unscaled" in model
        assert "scaled_var_unscaled_mu" in model
        assert "scaled_var_unscaled_sigma" in model

    def test_scaled_sample_prior(self) -> None:
        """Test that sample_prior works with the Scaled class."""
        normal = Prior("Normal", dims="channel")
        scaled = Scaled(normal, factor=2.0)

        coords = {"channel": ["A", "B", "C"]}
        prior = sample_prior(scaled, coords=coords, draws=25, name="scaled_var")

        assert isinstance(prior, xr.Dataset)
        assert prior.sizes == {"chain": 1, "draw": 25, "channel": 3}
        assert "scaled_var" in prior
        assert "scaled_var_unscaled" in prior


class TestCensored:
    def test_censored_is_variable_factory(
        self,
    ) -> None:
        normal = Prior("Normal")
        censored_normal = Censored(normal, lower=0)

        assert isinstance(censored_normal, VariableFactory)

    def test_deserialize_censored(self) -> None:
        data = {
            "class": "Censored",
            "data": {
                "dist": {
                    "dist": "Normal",
                },
                "lower": 0,
                "upper": float("inf"),
            },
        }

        instance = deserialize(data)
        assert isinstance(instance, Censored)
        assert isinstance(instance.distribution, Prior)
        assert instance.lower == 0
        assert instance.upper == float("inf")

    @pytest.mark.parametrize(
        "dims, expected_dims",
        [
            ("channel", ("channel",)),
            (("channel", "geo"), ("channel", "geo")),
        ],
        ids=["string", "tuple"],
    )
    def test_censored_dims_from_distribution(self, dims, expected_dims) -> None:
        normal = Prior("Normal", dims=dims)
        censored_normal = Censored(normal, lower=0)

        assert censored_normal.dims == expected_dims

    def test_censored_variables_created(
        self,
    ) -> None:
        normal = Prior("Normal", mu=Prior("Normal"), dims="dim")
        censored_normal = Censored(normal, lower=0)

        coords = {"dim": range(3)}
        with pm.Model(coords=coords) as model:
            censored_normal.create_variable("var")

        var_names = ["var", "var_mu"]
        assert set(var.name for var in model.unobserved_RVs) == set(var_names)
        dims = [(3,), ()]
        for var_name, dim in zip(var_names, dims, strict=False):
            assert fast_eval(model[var_name]).shape == dim

    def test_censored_sample_prior(
        self,
    ) -> None:
        normal = Prior("Normal", dims="channel")
        censored_normal = Censored(normal, lower=0)

        coords = {"channel": ["A", "B", "C"]}
        prior = censored_normal.sample_prior(coords=coords, draws=25)

        assert isinstance(prior, xr.Dataset)
        assert prior.sizes == {"chain": 1, "draw": 25, "channel": 3}

    def test_censored_to_graph(
        self,
    ) -> None:
        normal = Prior("Normal", dims="channel")
        censored_normal = Censored(normal, lower=0)

        G = censored_normal.to_graph()
        assert isinstance(G, Digraph)

    def test_censored_likelihood_variable(
        self,
    ) -> None:
        normal = Prior("Normal", sigma=Prior("HalfNormal"), dims="channel")
        censored_normal = Censored(normal, lower=0)

        coords = {"channel": range(3)}
        with pm.Model(coords=coords) as model:
            mu = pm.Normal("mu")
            variable = censored_normal.create_likelihood_variable(
                name="likelihood",
                mu=mu,
                observed=[1, 2, 3],
            )

        assert isinstance(variable, pt.TensorVariable)
        assert model.observed_RVs == [variable]
        assert "likelihood_sigma" in model

    def test_censored_likelihood_unsupported_distribution(
        self,
    ) -> None:
        cauchy = Prior("Cauchy")
        censored_cauchy = Censored(cauchy, lower=0)

        with pm.Model():
            mu = pm.Normal("mu")
            with pytest.raises(UnsupportedDistributionError):
                censored_cauchy.create_likelihood_variable(
                    name="likelihood",
                    mu=mu,
                    observed=1,
                )

    def test_censored_likelihood_already_has_mu(
        self,
    ) -> None:
        normal = Prior("Normal", mu=Prior("Normal"), sigma=Prior("HalfNormal"))
        censored_normal = Censored(normal, lower=0)

        with pm.Model():
            mu = pm.Normal("mu")
            with pytest.raises(MuAlreadyExistsError):
                censored_normal.create_likelihood_variable(
                    name="likelihood",
                    mu=mu,
                    observed=1,
                )

    def test_censored_to_dict(
        self,
    ) -> None:
        normal = Prior("Normal", mu=0, sigma=1, dims="channel")
        censored_normal = Censored(normal, lower=0)

        data = censored_normal.to_dict()
        assert data == {
            "class": "Censored",
            "data": {"dist": normal.to_dict(), "lower": 0, "upper": float("inf")},
        }

    @pytest.mark.parametrize(
        "mu",
        [
            0,
            np.arange(10),
        ],
        ids=["scalar", "vector"],
    )
    def test_censored_logp(self, mu) -> None:
        n_points = 10
        observed = np.zeros(n_points)
        coords = {"idx": range(n_points)}
        with pm.Model(coords=coords) as model:
            normal = Prior("Normal", dims="idx")
            Censored(normal, lower=0).create_likelihood_variable(
                "censored_normal",
                observed=observed,
                mu=mu,
            )
        logp = model.compile_logp()

        with pm.Model() as expected_model:
            pm.Censored(
                "censored_normal",
                pm.Normal.dist(mu=mu, sigma=1, shape=n_points),
                lower=0,
                upper=np.inf,
                observed=observed,
            )
        expected_logp = expected_model.compile_logp()

        point = {}
        np.testing.assert_allclose(logp(point), expected_logp(point))

    def test_censored_with_tensor_variable(self) -> None:
        normal = Prior("Normal", dims="channel")
        lower = pt.as_tensor_variable([0, 1, 2])
        censored_normal = Censored(normal, lower=lower)

        assert censored_normal.to_dict() == {
            "class": "Censored",
            "data": {
                "dist": normal.to_dict(),
                "lower": [0, 1, 2],
                "upper": float("inf"),
            },
        }

    def test_censored_dims_setter(self) -> None:
        normal = Prior("Normal", dims="channel")
        censored_normal = Censored(normal, lower=0)
        censored_normal.dims = "date"
        assert normal.dims == ("date",)

    def test_censored_normal(self) -> None:
        coords = {"idx": range(5)}
        observed = np.arange(5, dtype=float)
        mu = np.pi

        with pm.Model(coords=coords) as model:
            sigma = Prior("HalfNormal")
            normal = Prior("Normal", sigma=sigma, dims="idx")
            Censored(normal, lower=0).create_likelihood_variable(
                "censored_normal",
                mu=mu,
                observed=observed,
            )

        with pm.Model(coords=coords) as expected_model:
            sigma = pm.HalfNormal("censored_normal_sigma")
            normal = pm.Normal.dist(mu=mu, sigma=sigma)
            pm.Censored(
                "censored_normal",
                normal,
                lower=0,
                upper=np.inf,
                observed=observed,
                dims="idx",
            )

        # This doesn't work because of no OpFromGraph equality impl
        # assert equivalent_models(model, expected_model)

        ip = model.initial_point()
        np.testing.assert_allclose(model.compile_logp()(ip), expected_model.compile_logp()(ip))

    def test_censored_with_alternative_class(self) -> None:
        def is_type(data):
            return isinstance(data, dict) and "distribution" in data

        def deserialize_func(data):
            return Prior(**data)

        register_deserialization(is_type=is_type, deserialize=deserialize_func)

        data = {
            "class": "Censored",
            "data": {
                "dist": {
                    "distribution": "Normal",
                },
                "lower": 0,
                "upper": 10,
            },
        }

        instance = deserialize(data)

        assert isinstance(instance, Censored)
        assert instance.lower == 0
        assert instance.upper == 10
        assert instance.distribution == Prior("Normal")

        DESERIALIZERS.pop()


@pytest.mark.filterwarnings(
    "ignore:The `pymc.dims` module is experimental and may contain critical bugs"
)
class TestXDist:
    def test_xdist_serialization(self):
        import pymc.dims as pmd

        mu = pmd.as_xtensor([1, 2, 3], dims=("city",))
        sigma = DataArray([4, 5], dims=("country",))
        dims = ("city", "batch", "country")

        prior = Prior(
            "Normal",
            mu=mu,
            sigma=sigma,
            dims=dims,
        )

        data = prior.to_dict()
        assert data == {
            "dims": ("city", "batch", "country"),
            "dist": "Normal",
            "kwargs": {
                "mu": {
                    "class": "DataArray",
                    "data": [1, 2, 3],
                    "dims": ["city"],
                },
                "sigma": {
                    "class": "DataArray",
                    "data": [4, 5],
                    "dims": ["country"],
                },
            },
        }

        prior_again = deserialize(data)
        # Commented out because Prior equality fails with PyTensor / Xarray variables in the parameters
        # assert prior_again == prior

        data_again = prior_again.to_dict()
        assert data_again == data

    @pytest.mark.parametrize("transform", (None, "exp"))
    def test_xdist_prior(self, transform):
        import pymc.dims as pmd

        mu = pmd.as_xtensor([1, 2, 3], dims=("city",))
        sigma = DataArray([4, 5], dims=("country",))
        dims = ("city", "batch", "country")
        coords = {
            "city": range(3),
            "country": range(2),
            "batch": range(5),
        }

        prior = Prior(
            "Normal",
            mu=mu,
            sigma=sigma,
            dims=dims,
            transform=transform,
        )

        res = prior.sample_prior(draws=7, coords=coords, xdist=True)
        assert res.sizes == {"chain": 1, "draw": 7, "city": 3, "batch": 5, "country": 2}

        with pm.Model(coords=coords) as prior_m:
            prior.create_variable("x", xdist=True)

        if transform is None:
            with pm.Model(coords=coords) as expected_prior_m:
                pmd.Normal("x", mu=mu, sigma=sigma, dims=dims)
        else:
            with pm.Model(coords=coords) as expected_prior_m:
                x_raw = pmd.Normal("x_raw", mu=mu, sigma=sigma, dims=dims)
                pmd.Deterministic("x", pmd.math.exp(x_raw))

        assert equivalent_models(prior_m, expected_prior_m)

    def test_xdist_likelihood(self):
        import pymc.dims as pmd

        mu = pmd.as_xtensor([1, 2, 3], dims=("city",))
        sigma = DataArray([4, 5], dims=("country",))
        dims = ("city", "batch", "country")
        coords = {
            "batch": range(5),
            "city": range(3),
            "country": range(2),
        }

        likelihood = Prior(
            "Normal",
            sigma=sigma,
            dims=dims,
        )
        observed = np.random.normal(size=(3, 5, 2))
        with pm.Model(coords=coords) as obs_m:
            x_obs = pmd.Data("x_obs", observed, dims=dims)
            likelihood.create_likelihood_variable("x", mu=mu, observed=x_obs.T, xdist=True)

        with pm.Model(coords=coords) as expected_obs_m:
            x_obs = pmd.Data("x_obs", observed, dims=dims)
            pmd.Normal("x", mu=mu, sigma=sigma, observed=x_obs.T, dims=dims)

        assert equivalent_models(obs_m, expected_obs_m)

    def test_dims(self) -> None:
        assert Prior("Normal").dims is None
        assert Prior("Normal", dims=()).dims == ()

        p = Prior("Normal", mu=Prior("Normal", dims=("city",)))
        assert p.dims is None

        coords = {"city": range(3)}
        with pm.Model(coords=coords) as m:
            with pytest.raises(UnsupportedShapeError):
                p.create_variable("x")

        with pm.Model(coords=coords) as m:
            # xdist can infer dims on its own
            p.create_variable("x", xdist=True)
            assert m.named_vars_to_dims["x"] == ("city",)

        p.dims = ("city",)
        with pm.Model(coords=coords) as m:
            p.create_variable("x")
            assert m.named_vars_to_dims["x"] == ("city",)

        # This is always invalid
        with pytest.raises(UnsupportedShapeError):
            p.dims = ()



import numpy as np
import pytensor
import pytest

from pytensor import config
from pytensor.graph.traversal import explicit_graph_inputs

from pymc_extras.statespace.models import structural as st
from pymc_extras.statespace.models.structural.components.seasonality import FrequencySeasonality
from tests.statespace.models.structural.conftest import _assert_basic_coords_correct
from tests.statespace.test_utilities import assert_pattern_repeats, simulate_from_numpy_model

ATOL = 1e-8 if config.floatX.endswith("64") else 1e-4
RTOL = 0 if config.floatX.endswith("64") else 1e-6


@pytest.mark.parametrize("s", [10, 25, 50])
@pytest.mark.parametrize("d", [1, 3])
@pytest.mark.parametrize("innovations", [True, False])
@pytest.mark.parametrize("remove_first_state", [True, False])
@pytest.mark.filterwarnings(
    "ignore:divide by zero encountered in matmul:RuntimeWarning",
    "ignore:overflow encountered in matmul:RuntimeWarning",
    "ignore:invalid value encountered in matmul:RuntimeWarning",
)
def test_time_seasonality(s, d, innovations, remove_first_state, rng):
    def random_word(rng):
        return "".join(rng.choice(list("abcdefghijklmnopqrstuvwxyz")) for _ in range(5))

    state_names = tuple(random_word(rng) for _ in range(s * d))
    mod = st.TimeSeasonality(
        season_length=s,
        duration=d,
        innovations=innovations,
        name="season",
        state_names=state_names,
        remove_first_state=remove_first_state,
    )
    x0 = np.zeros(mod.k_states // mod.duration, dtype=config.floatX)
    x0[0] = 1

    params = {"params_season": x0}
    if innovations:
        params["sigma_season"] = 0.0

    x, y = simulate_from_numpy_model(mod, rng, params, steps=100 * mod.duration)
    y = y.ravel()
    if not innovations:
        assert_pattern_repeats(y, s * d, atol=ATOL, rtol=RTOL)

    # Check coords
    mod = mod.build(verbose=False)
    _assert_basic_coords_correct(mod)

    test_slice = slice(d, None) if remove_first_state else slice(None)
    assert mod.coords["state_season"] == state_names[test_slice]


@pytest.mark.parametrize("d", [1, 3])
@pytest.mark.parametrize(
    "remove_first_state", [True, False], ids=["remove_first_state", "keep_first_state"]
)
def test_time_seasonality_multiple_observed(rng, d, remove_first_state):
    s = 3
    state_names = [f"state_{i}_{j}" for i in range(s) for j in range(d)]
    mod = st.TimeSeasonality(
        season_length=s,
        duration=d,
        innovations=True,
        name="season",
        state_names=state_names,
        observed_state_names=["data_1", "data_2"],
        remove_first_state=remove_first_state,
    )
    x0 = np.zeros((mod.k_endog, mod.k_states // mod.k_endog // mod.duration), dtype=config.floatX)

    expected_states = [
        f"state_{i}_{j}[data_{k}]"
        for k in range(1, 3)
        for i in range(int(remove_first_state), s)
        for j in range(d)
    ]
    assert mod.state_names == tuple(expected_states)
    assert mod.shock_names == ("season[data_1]", "season[data_2]")

    x0[0, 0] = 1
    x0[1, 0] = 2.0

    params = {"params_season": x0, "sigma_season": np.array([0.0, 0.0], dtype=config.floatX)}

    x, y = simulate_from_numpy_model(mod, rng, params, steps=123 * d)
    assert_pattern_repeats(y[:, 0], s * d, atol=ATOL, rtol=RTOL)
    assert_pattern_repeats(y[:, 1], s * d, atol=ATOL, rtol=RTOL)

    mod = mod.build(verbose=False)
    x0, *_, T, Z, R, _, Q = mod._unpack_statespace_with_placeholders()

    input_vars = explicit_graph_inputs([x0, T, Z, R, Q])

    fn = pytensor.function(
        inputs=list(input_vars),
        outputs=[x0, T, Z, R, Q],
        mode="FAST_COMPILE",
    )

    params["sigma_season"] = np.array([0.1, 0.8], dtype=config.floatX)
    x0, T, Z, R, Q = fn(**params)

    # Because the dimension of the observed states is 2,
    # the expected T is the diagonal block matrix [[T0, 0], [0, T0]]
    # where T0 is the transition matrix we would have if the
    # seasonality were not multiple observed.
    mod0 = st.TimeSeasonality(season_length=s, duration=d, remove_first_state=remove_first_state)
    T0 = mod0.ssm["transition"].eval()

    if remove_first_state:
        expected_x0 = np.repeat(np.array([1.0, 0.0, 2.0, 0.0]), d)
        expected_T = np.block(
            [[T0, np.zeros((d * (s - 1), d * (s - 1)))], [np.zeros((d * (s - 1), d * (s - 1))), T0]]
        )
        expected_R = np.array(
            [[1.0, 1.0]] + [[0.0, 0.0]] * (2 * d - 1) + [[1.0, 1.0]] + [[0.0, 0.0]] * (2 * d - 1)
        )
        Z0 = np.zeros((2, d * (s - 1)))
        Z0[0, 0] = 1
        Z1 = np.zeros((2, d * (s - 1)))
        Z1[1, 0] = 1
        expected_Z = np.block([[Z0, Z1]])

    else:
        expected_x0 = np.repeat(np.array([1.0, 0.0, 0.0, 2.0, 0.0, 0.0]), d)
        expected_T = np.block([[T0, np.zeros((s * d, s * d))], [np.zeros((s * d, s * d)), T0]])
        expected_R = np.array(
            [[1.0, 1.0]] + [[0.0, 0.0]] * (s * d - 1) + [[1.0, 1.0]] + [[0.0, 0.0]] * (s * d - 1)
        )
        Z0 = np.zeros((2, s * d))
        Z0[0, 0] = 1
        Z1 = np.zeros((2, s * d))
        Z1[1, 0] = 1
        expected_Z = np.block([[Z0, Z1]])

    expected_Q = np.array([[0.1**2, 0.0], [0.0, 0.8**2]])

    for matrix, expected in zip(
        [x0, T, Z, R, Q],
        [expected_x0, expected_T, expected_Z, expected_R, expected_Q],
    ):
        np.testing.assert_allclose(matrix, expected)


def test_time_seasonality_shared_states():
    mod = st.TimeSeasonality(
        season_length=3,
        duration=1,
        innovations=True,
        name="season",
        state_names=["season_1", "season_2", "season_3"],
        observed_state_names=["data_1", "data_2"],
        remove_first_state=False,
        share_states=True,
    )

    assert mod.k_endog == 2
    assert mod.k_states == 3
    assert mod.k_posdef == 1

    assert mod.coords["state_season"] == ("season_1", "season_2", "season_3")

    assert mod.state_names == (
        "season_1[season_shared]",
        "season_2[season_shared]",
        "season_3[season_shared]",
    )
    assert mod.shock_names == ("season[shared]",)

    Z, T, R = pytensor.function(
        [], [mod.ssm["design"], mod.ssm["transition"], mod.ssm["selection"]], mode="FAST_COMPILE"
    )()

    np.testing.assert_allclose(np.array([[1.0, 0.0, 0.0], [1.0, 0.0, 0.0]]), Z)

    np.testing.assert_allclose(np.array([[0.0, 1.0, 0.0], [0.0, 0.0, 1.0], [1.0, 0.0, 0.0]]), T)

    np.testing.assert_allclose(np.array([[1.0], [0.0], [0.0]]), R)


def test_add_mixed_shared_not_shared_time_seasonality():
    shared_season = st.TimeSeasonality(
        season_length=3,
        duration=1,
        innovations=True,
        name="shared",
        state_names=["season_1", "season_2", "season_3"],
        observed_state_names=["data_1", "data_2"],
        remove_first_state=False,
        share_states=True,
    )
    individual_season = st.TimeSeasonality(
        season_length=3,
        duration=1,
        innovations=False,
        name="individual",
        state_names=["season_1", "season_2", "season_3"],
        observed_state_names=["data_1", "data_2"],
        remove_first_state=True,
        share_states=False,
    )
    mod = (shared_season + individual_season).build(verbose=False)

    assert mod.k_endog == 2
    assert mod.k_states == 7
    assert mod.k_posdef == 1

    assert mod.coords["state_shared"] == ("season_1", "season_2", "season_3")
    assert mod.coords["state_individual"] == ("season_2", "season_3")

    assert mod.state_names == (
        "season_1[shared_shared]",
        "season_2[shared_shared]",
        "season_3[shared_shared]",
        "season_2[data_1]",
        "season_3[data_1]",
        "season_2[data_2]",
        "season_3[data_2]",
    )

    Z, T, R = pytensor.function(
        [], [mod.ssm["design"], mod.ssm["transition"], mod.ssm["selection"]], mode="FAST_COMPILE"
    )()

    np.testing.assert_allclose(
        np.array([[1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0], [1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0]]), Z
    )

    np.testing.assert_allclose(
        np.array(
            [
                [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, -1.0, -1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, -1.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
            ]
        ),
        T,
    )

    np.testing.assert_allclose(np.array([[1.0], [0.0], [0.0], [0.0], [0.0], [0.0], [0.0]]), R)


@pytest.mark.parametrize("d1, d2", [(1, 1), (1, 3), (3, 1), (3, 3)])
def test_add_two_time_seasonality_different_observed(rng, d1, d2):
    mod1 = st.TimeSeasonality(
        season_length=3,
        duration=d1,
        innovations=True,
        name="season1",
        state_names=[f"state_{i}_{j}" for i in range(3) for j in range(d1)],
        observed_state_names=["data_1"],
        remove_first_state=False,
    )
    mod2 = st.TimeSeasonality(
        season_length=5,
        duration=d2,
        innovations=True,
        name="season2",
        state_names=[f"state_{i}_{j}" for i in range(5) for j in range(d2)],
        observed_state_names=["data_2"],
    )

    mod = (mod1 + mod2).build(verbose=False)

    params = {
        "params_season1": np.array([1.0, 0.0, 0.0], dtype=config.floatX),
        "params_season2": np.array([3.0, 0.0, 0.0, 0.0], dtype=config.floatX),
        "sigma_season1": np.array(0.0, dtype=config.floatX),
        "sigma_season2": np.array(0.0, dtype=config.floatX),
        "initial_state_cov": np.eye(mod.k_states, dtype=config.floatX),
    }

    x, y = simulate_from_numpy_model(mod, rng, params, steps=3 * 5 * 5 * d1 * d2)
    assert_pattern_repeats(y[:, 0], 3 * d1, atol=ATOL, rtol=RTOL)
    assert_pattern_repeats(y[:, 1], 5 * d2, atol=ATOL, rtol=RTOL)

    assert mod.state_names == tuple(
        [
            item
            for sublist in [
                [f"state_0_{j}[data_1]" for j in range(d1)],
                [f"state_1_{j}[data_1]" for j in range(d1)],
                [f"state_2_{j}[data_1]" for j in range(d1)],
                [f"state_1_{j}[data_2]" for j in range(d2)],
                [f"state_2_{j}[data_2]" for j in range(d2)],
                [f"state_3_{j}[data_2]" for j in range(d2)],
                [f"state_4_{j}[data_2]" for j in range(d2)],
            ]
            for item in sublist
        ]
    )

    assert mod.shock_names == (
        "season1[data_1]",
        "season2[data_2]",
    )

    x0, *_, T = mod._unpack_statespace_with_placeholders()[:5]
    input_vars = explicit_graph_inputs([x0, T])
    fn = pytensor.function(
        inputs=list(input_vars),
        outputs=[x0, T],
        mode="FAST_COMPILE",
    )

    x0, T = fn(
        params_season1=np.array([1.0, 0.0, 0.0], dtype=config.floatX),
        params_season2=np.array([3.0, 0.0, 0.0, 1.2], dtype=config.floatX),
    )

    np.testing.assert_allclose(
        np.repeat(np.array([1.0, 0.0, 0.0, 3.0, 0.0, 0.0, 1.2]), [d1, d1, d1, d2, d2, d2, d2]),
        x0,
        atol=ATOL,
        rtol=RTOL,
    )

    # The transition matrix T of mod is expected to be [[T1, 0], [0, T2]],
    # where T1 and T2 are the transition matrices of mod1 and mod2, respectively.
    T1 = mod1.ssm["transition"].eval()
    T2 = mod2.ssm["transition"].eval()
    np.testing.assert_allclose(
        np.block(
            [[T1, np.zeros((T1.shape[0], T2.shape[1]))], [np.zeros((T2.shape[0], T1.shape[1])), T2]]
        ),
        T,
        atol=ATOL,
        rtol=RTOL,
    )


def get_shift_factor(s):
    s_str = str(s)
    if "." not in s_str:
        return 1
    _, decimal = s_str.split(".")
    return 10 ** len(decimal)


@pytest.mark.parametrize("n", [*np.arange(1, 6, dtype="int").tolist(), None])
@pytest.mark.parametrize("s", [5, 10, 25, 25.2])
def test_frequency_seasonality(n, s, rng):
    mod = st.FrequencySeasonality(season_length=s, n=n, name="season")
    assert mod.param_info["sigma_season"].shape == ()  # scalar for univariate
    assert mod.param_info["sigma_season"].dims is None
    assert len(mod.coords["state_season"]) == mod.n_coefs

    x0 = rng.normal(size=mod.n_coefs).astype(config.floatX)
    params = {"params_season": x0, "sigma_season": 0.0}
    k = get_shift_factor(s)
    T = int(s * k)

    x, y = simulate_from_numpy_model(mod, rng, params, steps=2 * T)
    assert_pattern_repeats(y, T, atol=ATOL, rtol=RTOL)

    # check coords
    mod = mod.build(verbose=False)
    _assert_basic_coords_correct(mod)

    if n is None:
        n = int(s // 2)
    states = [f"{f}_{i}_season" for i in range(n) for f in ["Cos", "Sin"]]

    # remove last state when model is completely saturated
    if s / n == 2.0:
        states.pop()
    assert mod.coords["state_season"] == tuple(states)


def test_frequency_seasonality_multiple_observed(rng):
    observed_state_names = ["data_1", "data_2"]
    season_length = 4
    mod = st.FrequencySeasonality(
        season_length=season_length,
        n=None,
        name="season",
        innovations=True,
        observed_state_names=observed_state_names,
    )
    assert mod.param_info["params_season"].shape == (mod.k_endog, mod.n_coefs)
    assert mod.param_info["params_season"].dims == ("endog_season", "state_season")
    assert mod.param_info["sigma_season"].dims == ("endog_season",)

    expected_state_names = (
        "Cos_0_season[data_1]",
        "Sin_0_season[data_1]",
        "Cos_1_season[data_1]",
        "Sin_1_season[data_1]",
        "Cos_0_season[data_2]",
        "Sin_0_season[data_2]",
        "Cos_1_season[data_2]",
        "Sin_1_season[data_2]",
    )
    assert mod.state_names == expected_state_names
    assert mod.shock_names == (
        "Cos_0_season[data_1]",
        "Sin_0_season[data_1]",
        "Cos_1_season[data_1]",
        "Sin_1_season[data_1]",
        "Cos_0_season[data_2]",
        "Sin_0_season[data_2]",
        "Cos_1_season[data_2]",
        "Sin_1_season[data_2]",
    )

    x0 = np.zeros((2, 3), dtype=config.floatX)
    x0[0, 0] = 1.0
    x0[1, 0] = 2.0
    params = {"params_season": x0, "sigma_season": np.zeros(2, dtype=config.floatX)}
    x, y = simulate_from_numpy_model(mod, rng, params, steps=12)

    # check periodicity for each observed series
    assert_pattern_repeats(y[:, 0], 4, atol=ATOL, rtol=RTOL)
    assert_pattern_repeats(y[:, 1], 4, atol=ATOL, rtol=RTOL)

    mod = mod.build(verbose=False)
    assert mod.coords["state_season"] == (
        "Cos_0_season",
        "Sin_0_season",
        "Cos_1_season",
    )

    x0_sym, *_, T_sym, Z_sym, R_sym, _, Q_sym = mod._unpack_statespace_with_placeholders()
    input_vars = explicit_graph_inputs([x0_sym, T_sym, Z_sym, R_sym, Q_sym])
    fn = pytensor.function(
        inputs=list(input_vars),
        outputs=[x0_sym, T_sym, Z_sym, R_sym, Q_sym],
        mode="FAST_COMPILE",
    )
    params["sigma_season"] = np.array([0.1, 0.8], dtype=config.floatX)
    x0_v, T_v, Z_v, R_v, Q_v = fn(**params)

    # x0 should be raveled into a single vector, with data_1 states first, then data_2 states
    np.testing.assert_allclose(
        x0_v, np.array([1.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0, 0.0]), atol=ATOL, rtol=RTOL
    )

    # T_v shape: (8, 8) (k_endog * k_states)
    # The transition matrix is block diagonal, each block is:
    # For n=2, season_length=4:
    # lambda_1 = 2*pi*1/4 = pi/2, cos(pi/2)=0, sin(pi/2)=1
    # lambda_2 = 2*pi*2/4 = pi,   cos(pi)=-1, sin(pi)=0
    # Block 1 (Cos_0, Sin_0):
    # [[cos(pi/2), sin(pi/2)],
    #  [-sin(pi/2), cos(pi/2)]] = [[0, 1], [-1, 0]]
    # Block 2 (Cos_1, Sin_1):
    # [[-1, 0], [0, -1]]
    expected_T_block1 = np.array([[0.0, 1.0], [-1.0, 0.0]])
    expected_T_block2 = np.array([[-1.0, 0.0], [0.0, -1.0]])
    expected_T = np.zeros((8, 8))
    # data_1
    expected_T[0:2, 0:2] = expected_T_block1
    expected_T[2:4, 2:4] = expected_T_block2
    # data_2
    expected_T[4:6, 4:6] = expected_T_block1
    expected_T[6:8, 6:8] = expected_T_block2
    np.testing.assert_allclose(T_v, expected_T, atol=ATOL, rtol=RTOL)

    # Only the first two states (one sin and one cos component) of each observed series are observed
    expected_Z = np.zeros((2, 8))
    expected_Z[0, 0] = 1.0
    expected_Z[0, 2] = 1.0
    expected_Z[1, 4] = 1.0
    expected_Z[1, 6] = 1.0
    np.testing.assert_allclose(Z_v, expected_Z, atol=ATOL, rtol=RTOL)

    np.testing.assert_allclose(R_v, np.eye(8), atol=ATOL, rtol=RTOL)

    Q_diag = np.diag(Q_v)
    expected_Q_diag = np.r_[np.full(4, 0.1**2), np.full(4, 0.8**2)]
    np.testing.assert_allclose(Q_diag, expected_Q_diag, atol=ATOL, rtol=RTOL)


def test_frequency_seasonality_multivariate_shared_states():
    mod = st.FrequencySeasonality(
        season_length=4,
        n=1,
        name="season",
        innovations=True,
        observed_state_names=["data_1", "data_2"],
        share_states=True,
    )

    assert mod.k_endog == 2
    assert mod.k_states == 2
    assert mod.k_posdef == 2

    assert mod.state_names == (
        "Cos_0_season[shared]",
        "Sin_0_season[shared]",
    )
    assert mod.shock_names == (
        "Cos_0_season[shared]",
        "Sin_0_season[shared]",
    )

    assert mod.coords["state_season"] == ("Cos_0_season", "Sin_0_season")

    Z, T, R = pytensor.function(
        [], [mod.ssm["design"], mod.ssm["transition"], mod.ssm["selection"]], mode="FAST_COMPILE"
    )()

    np.testing.assert_allclose(np.array([[1.0, 0.0], [1.0, 0.0]]), Z)

    np.testing.assert_allclose(np.array([[1.0, 0.0], [0.0, 1.0]]), R)

    lam = 2 * np.pi * 1 / 4
    np.testing.assert_allclose(
        np.array([[np.cos(lam), np.sin(lam)], [-np.sin(lam), np.cos(lam)]]), T
    )


def test_add_two_frequency_seasonality_different_observed(rng):
    mod1 = st.FrequencySeasonality(
        season_length=4,
        n=2,  # saturated
        name="freq1",
        innovations=True,
        observed_state_names=["data_1"],
    )
    mod2 = st.FrequencySeasonality(
        season_length=6,
        n=1,  # unsaturated
        name="freq2",
        innovations=True,
        observed_state_names=["data_2"],
    )

    mod = (mod1 + mod2).build(verbose=False)

    params = {
        "params_freq1": np.array([1.0, 0.0, 0.0], dtype=config.floatX),
        "params_freq2": np.array([3.0, 0.0], dtype=config.floatX),
        "sigma_freq1": np.array(0.0, dtype=config.floatX),
        "sigma_freq2": np.array(0.0, dtype=config.floatX),
        "initial_state_cov": np.eye(mod.k_states, dtype=config.floatX),
    }

    x, y = simulate_from_numpy_model(mod, rng, params, steps=4 * 6 * 3)

    assert_pattern_repeats(y[:, 0], 4, atol=ATOL, rtol=RTOL)
    assert_pattern_repeats(y[:, 1], 6, atol=ATOL, rtol=RTOL)

    assert mod.state_names == (
        "Cos_0_freq1[data_1]",
        "Sin_0_freq1[data_1]",
        "Cos_1_freq1[data_1]",
        "Sin_1_freq1[data_1]",
        "Cos_0_freq2[data_2]",
        "Sin_0_freq2[data_2]",
    )

    assert mod.shock_names == (
        "Cos_0_freq1[data_1]",
        "Sin_0_freq1[data_1]",
        "Cos_1_freq1[data_1]",
        "Sin_1_freq1[data_1]",
        "Cos_0_freq2[data_2]",
        "Sin_0_freq2[data_2]",
    )

    x0, *_, T = mod._unpack_statespace_with_placeholders()[:5]
    input_vars = explicit_graph_inputs([x0, T])
    fn = pytensor.function(
        inputs=list(input_vars),
        outputs=[x0, T],
        mode="FAST_COMPILE",
    )

    x0_v, T_v = fn(
        params_freq1=np.array([1.0, 0.0, 1.2], dtype=config.floatX),
        params_freq2=np.array([3.0, 0.0], dtype=config.floatX),
    )

    # Make sure the extra 0 in from the first component (the saturated state) is there!
    np.testing.assert_allclose(np.array([1.0, 0.0, 1.2, 0.0, 3.0, 0.0]), x0_v, atol=ATOL, rtol=RTOL)

    # Transition matrix is block diagonal: 4x4 for freq1, 2x2 for freq2
    # freq1: n=4, lambdas = 2*pi*1/6, 2*pi*2/6
    lam1 = 2 * np.pi * 1 / 4
    lam2 = 2 * np.pi * 2 / 4
    freq1_T1 = np.array([[np.cos(lam1), np.sin(lam1)], [-np.sin(lam1), np.cos(lam1)]])
    freq1_T2 = np.array([[np.cos(lam2), np.sin(lam2)], [-np.sin(lam2), np.cos(lam2)]])
    freq1_T = np.zeros((4, 4))

    # freq2: n=4, lambdas = 2*pi*1/6
    lam3 = 2 * np.pi * 1 / 6
    freq2_T = np.array([[np.cos(lam3), np.sin(lam3)], [-np.sin(lam3), np.cos(lam3)]])

    freq1_T[0:2, 0:2] = freq1_T1
    freq1_T[2:4, 2:4] = freq1_T2

    expected_T = np.zeros((6, 6))
    expected_T[0:4, 0:4] = freq1_T
    expected_T[4:6, 4:6] = freq2_T

    np.testing.assert_allclose(expected_T, T_v, atol=ATOL, rtol=RTOL)


def test_add_frequency_seasonality_shared_and_not_shared():
    shared_season = st.FrequencySeasonality(
        season_length=4,
        n=1,
        name="shared_season",
        innovations=True,
        observed_state_names=["data_1", "data_2"],
        share_states=True,
    )

    individual_season = st.FrequencySeasonality(
        season_length=4,
        n=2,
        name="individual_season",
        innovations=True,
        observed_state_names=["data_1", "data_2"],
        share_states=False,
    )

    mod = (shared_season + individual_season).build(verbose=False)

    assert mod.k_endog == 2
    assert mod.k_states == 10
    assert mod.k_posdef == 10

    assert mod.coords["state_shared_season"] == (
        "Cos_0_shared_season",
        "Sin_0_shared_season",
    )
    assert mod.coords["state_individual_season"] == (
        "Cos_0_individual_season",
        "Sin_0_individual_season",
        "Cos_1_individual_season",
    )


@pytest.mark.parametrize(
    "test_case",
    [
        {
            "name": "single_endog_non_saturated",
            "season_length": 12,
            "n": 2,
            "observed_state_names": ("data1",),
            "expected_shape": (4,),
        },
        {
            "name": "single_endog_saturated",
            "season_length": 12,
            "n": 6,
            "observed_state_names": ("data1",),
            "expected_shape": (11,),
        },
        {
            "name": "multiple_endog_non_saturated",
            "season_length": 12,
            "n": 2,
            "observed_state_names": (
                "data1",
                "data2",
            ),
            "expected_shape": (2, 4),
        },
        {
            "name": "multiple_endog_saturated",
            "season_length": 12,
            "n": 6,
            "observed_state_names": (
                "data1",
                "data2",
            ),
            "expected_shape": (2, 11),
        },
        {
            "name": "small_n",
            "season_length": 12,
            "n": 1,
            "observed_state_names": ("data1",),
            "expected_shape": (2,),
        },
        {
            "name": "many_endog",
            "season_length": 12,
            "n": 2,
            "observed_state_names": ("data1", "data2", "data3", "data4"),
            "expected_shape": (4, 4),
        },
    ],
    ids=lambda x: x["name"],
)
def test_frequency_seasonality_coordinates(test_case):
    model_name = f"season_{test_case['name'].split('_')[0]}"

    season = FrequencySeasonality(
        season_length=test_case["season_length"],
        n=test_case["n"],
        name=model_name,
        observed_state_names=test_case["observed_state_names"],
    )
    season.populate_component_properties()

    # assert parameter shape
    assert season.param_info[f"params_{model_name}"].shape == test_case["expected_shape"]

    # generate expected state names based on actual model name
    expected_state_names = tuple(
        [f"{f}_{i}_{model_name}" for i in range(test_case["n"]) for f in ["Cos", "Sin"]][
            : test_case["expected_shape"][-1]
        ]
    )

    # assert coordinate structure
    if len(test_case["observed_state_names"]) == 1:
        assert len(season.coords[f"state_{model_name}"]) == test_case["expected_shape"][0]
        assert season.coords[f"state_{model_name}"] == expected_state_names
    else:
        assert len(season.coords[f"endog_{model_name}"]) == test_case["expected_shape"][0]
        assert len(season.coords[f"state_{model_name}"]) == test_case["expected_shape"][1]
        assert season.coords[f"state_{model_name}"] == expected_state_names

    # Check coords match the expected shape
    param_shape = season.param_info[f"params_{model_name}"].shape
    state_coords = season.coords[f"state_{model_name}"]
    endog_coords = season.coords.get(f"endog_{model_name}")

    assert len(state_coords) == param_shape[-1]
    if endog_coords:
        assert len(endog_coords) == param_shape[0]



from typing import Protocol

import numpy as np

from pymc import Model, compile
from pymc.pytensorf import rewrite_pregrad
from pytensor import tensor as pt

from pymc_extras.inference.advi.autoguide import AutoGuideModel
from pymc_extras.inference.advi.objective import advi_objective, get_logp_logq
from pymc_extras.inference.advi.pytensorf import vectorize_random_graph


class TrainingFn(Protocol):
    def __call__(self, draws: int, *params: np.ndarray) -> tuple[np.ndarray, ...]: ...


def compile_svi_training_fn(
    model: Model, guide: AutoGuideModel, stick_the_landing: bool = True, **compile_kwargs
) -> TrainingFn:
    draws = pt.scalar("draws", dtype=int)
    params = guide.params

    logp, logq = get_logp_logq(model, guide, stick_the_landing=stick_the_landing)

    scalar_negative_elbo = advi_objective(logp, logq)
    [negative_elbo_draws] = vectorize_random_graph([scalar_negative_elbo], batch_draws=draws)
    negative_elbo = negative_elbo_draws.mean(axis=0)

    negative_elbo_grads = pt.grad(rewrite_pregrad(negative_elbo), wrt=params)

    if "trust_input" not in compile_kwargs:
        compile_kwargs["trust_input"] = True

    f_loss_dloss = compile(
        inputs=[draws, *params], outputs=[negative_elbo, *negative_elbo_grads], **compile_kwargs
    )

    return f_loss_dloss



"""Tests for pathfinder InferenceData integration."""

from collections import Counter
from dataclasses import dataclass

import numpy as np
import pytest
import xarray as xr

# Mock objects for testing without full dependencies
from pymc_extras.inference.pathfinder.lbfgs import LBFGSStatus
from pymc_extras.inference.pathfinder.pathfinder import PathfinderConfig, PathStatus


@dataclass
class MockPathfinderResult:
    """Mock PathfinderResult for testing."""

    samples: np.ndarray = None
    logP: np.ndarray = None
    logQ: np.ndarray = None
    lbfgs_niter: np.ndarray = None
    elbo_argmax: np.ndarray = None
    lbfgs_status: LBFGSStatus = LBFGSStatus.CONVERGED
    path_status: PathStatus = PathStatus.SUCCESS


@dataclass
class MockMultiPathfinderResult:
    """Mock MultiPathfinderResult for testing."""

    samples: np.ndarray = None
    logP: np.ndarray = None
    logQ: np.ndarray = None
    lbfgs_niter: np.ndarray = None
    elbo_argmax: np.ndarray = None
    lbfgs_status: Counter = None
    path_status: Counter = None
    importance_sampling: str = "psis"
    warnings: list = None
    pareto_k: float = None
    num_paths: int = None
    num_draws: int = None
    pathfinder_config: PathfinderConfig = None
    compile_time: float = None
    compute_time: float = None
    all_paths_failed: bool = False

    def __post_init__(self):
        if self.lbfgs_status is None:
            self.lbfgs_status = Counter()
        if self.path_status is None:
            self.path_status = Counter()
        if self.warnings is None:
            self.warnings = []


class TestPathfinderResultToXarray:
    """Tests for converting single PathfinderResult to xarray."""

    def test_single_result_basic_conversion(self):
        """Test basic conversion of PathfinderResult to xarray Dataset."""
        # Skip if dependencies not available
        pytest.importorskip("arviz")

        from pymc_extras.inference.pathfinder.idata import pathfinder_result_to_xarray

        # Create mock result
        result = MockPathfinderResult(
            samples=np.random.normal(0, 1, (1, 100, 2)),
            logP=np.random.normal(-10, 1, (1, 100)),
            logQ=np.random.normal(-11, 1, (1, 100)),
            lbfgs_niter=np.array([50]),
            elbo_argmax=np.array([25]),
        )

        ds = pathfinder_result_to_xarray(result, model=None)

        # Check basic structure
        assert isinstance(ds, xr.Dataset)
        assert "lbfgs_niter" in ds.data_vars
        assert "elbo_argmax" in ds.data_vars
        assert "lbfgs_status_code" in ds.data_vars
        assert "lbfgs_status_name" in ds.data_vars
        assert "path_status_code" in ds.data_vars
        assert "path_status_name" in ds.data_vars

        # Check attributes
        assert "lbfgs_status" in ds.attrs
        assert "path_status" in ds.attrs

    def test_parameter_coordinates(self):
        """Test parameter coordinate generation."""
        pytest.importorskip("arviz")

        from pymc_extras.inference.pathfinder.idata import get_param_coords

        # Test fallback to indices when no model
        coords = get_param_coords(None, 3)
        assert coords == ["0", "1", "2"]

    def test_get_param_coords_fail_fast(self):
        """Test that get_param_coords fails fast on model errors."""
        pytest.importorskip("arviz")
        pytest.importorskip("pymc")

        import pymc as pm

        from pymc_extras.inference.pathfinder.idata import get_param_coords

        # Test that it fails when model.initial_point() raises an exception
        with pm.Model() as broken_model:
            # Shape mismatch causes initial_point to fail
            x = pm.Normal("x", mu=[0, 1], sigma=1, shape=1)  # incompatible shapes

        with pytest.raises(ValueError, match=r".*incompatible.*"):
            get_param_coords(broken_model, 2)

        # Test that it works correctly with valid models
        with pm.Model() as valid_model:
            x = pm.Normal("x", 0, 1)  # scalar
            y = pm.Normal("y", 0, 1, shape=2)  # vector

        coords = get_param_coords(valid_model, 3)
        expected = ["x", "y[0]", "y[1]"]
        assert coords == expected

    def test_multipath_coordinate_dimensions_with_importance_sampling(self):
        """Test that path dimensions are calculated correctly when importance sampling collapses samples."""
        pytest.importorskip("arviz")

        from pymc_extras.inference.pathfinder.idata import multipathfinder_result_to_xarray

        # Mock a multi-path result where importance sampling has collapsed the samples
        # but per-path diagnostics are still available
        result = MockMultiPathfinderResult(
            samples=np.random.normal(0, 1, (1000, 2)),  # Collapsed: (total_draws, n_params)
            lbfgs_niter=np.array([50, 45, 55, 40]),  # Per-path: (4,)
            elbo_argmax=np.array([25, 30, 20, 35]),  # Per-path: (4,)
            logP=np.random.normal(-10, 1, (4, 250)),  # Per-path, per-draw: (4, 250)
            logQ=np.random.normal(-11, 1, (4, 250)),  # Per-path, per-draw: (4, 250)
            lbfgs_status=Counter({LBFGSStatus.CONVERGED: 4}),
            path_status=Counter({PathStatus.SUCCESS: 4}),
            num_paths=4,
            num_draws=1000,
        )

        ds = multipathfinder_result_to_xarray(result, model=None)

        # Check that path dimension is correctly inferred as 4 (not 1000)
        assert "path" in ds.dims
        assert ds.sizes["path"] == 4  # Should be 4 paths, not 1000 samples

        # Check that per-path data has correct shape with paths/ prefix
        assert "paths/lbfgs_niter" in ds.data_vars
        assert ds["paths/lbfgs_niter"].shape == (4,)
        assert "paths/elbo_argmax" in ds.data_vars
        assert ds["paths/elbo_argmax"].shape == (4,)

    def test_determine_num_paths_helper(self):
        """Test the _determine_num_paths helper function."""
        pytest.importorskip("arviz")

        from pymc_extras.inference.pathfinder.idata import _determine_num_paths

        # Test with lbfgs_niter
        result1 = MockMultiPathfinderResult(
            lbfgs_niter=np.array([10, 15, 12]),
            elbo_argmax=None,
        )
        assert _determine_num_paths(result1) == 3

        # Test with logP when lbfgs_niter is None
        result2 = MockMultiPathfinderResult(
            lbfgs_niter=None,
            logP=np.random.normal(0, 1, (5, 100)),  # 5 paths, 100 samples each
        )
        assert _determine_num_paths(result2) == 5

        # Test fallback to status counters
        result3 = MockMultiPathfinderResult(
            lbfgs_niter=None,
            elbo_argmax=None,
            logP=None,
            logQ=None,
            lbfgs_status=Counter({LBFGSStatus.CONVERGED: 2}),
        )
        assert _determine_num_paths(result3) == 2

    def test_status_counter_conversion(self):
        """Test conversion of status counters to DataArray."""
        pytest.importorskip("arviz")

        from pymc_extras.inference.pathfinder.idata import _status_counter_to_dataarray

        counter = Counter({LBFGSStatus.CONVERGED: 2, LBFGSStatus.MAX_ITER_REACHED: 1})
        da = _status_counter_to_dataarray(counter, LBFGSStatus)

        assert isinstance(da, xr.DataArray)
        assert "status" in da.dims
        assert da.sel(status="CONVERGED").item() == 2
        assert da.sel(status="MAX_ITER_REACHED").item() == 1


class TestMultiPathfinderResultToXarray:
    """Tests for converting MultiPathfinderResult to xarray."""

    def test_multi_result_conversion(self):
        """Test conversion of MultiPathfinderResult to consolidated dataset."""
        pytest.importorskip("arviz")

        from pymc_extras.inference.pathfinder.idata import multipathfinder_result_to_xarray
        from pymc_extras.inference.pathfinder.pathfinder import PathfinderConfig

        # Create mock config
        config = PathfinderConfig(
            num_draws=100,
            maxcor=5,
            maxiter=1000,
            ftol=1e-5,
            gtol=1e-8,
            maxls=1000,
            jitter=2.0,
            epsilon=1e-8,
            num_elbo_draws=10,
        )

        # Create mock multi-path result
        result = MockMultiPathfinderResult(
            samples=np.random.normal(0, 1, (3, 100, 2)),  # 3 paths, 100 draws, 2 params
            logP=np.random.normal(-10, 1, (3, 100)),
            logQ=np.random.normal(-11, 1, (3, 100)),
            lbfgs_niter=np.array([50, 45, 55]),
            elbo_argmax=np.array([25, 30, 20]),
            lbfgs_status=Counter({LBFGSStatus.CONVERGED: 3}),
            path_status=Counter({PathStatus.SUCCESS: 3}),
            num_paths=3,
            num_draws=300,
            compile_time=1.5,
            compute_time=10.2,
            pareto_k=0.5,
            pathfinder_config=config,
        )

        # Test without diagnostics
        ds = multipathfinder_result_to_xarray(result, model=None, store_diagnostics=False)

        # Check that we get a single consolidated dataset
        assert isinstance(ds, xr.Dataset)

        # Check summary data (top level)
        assert "num_paths" in ds.data_vars
        assert "num_draws" in ds.data_vars
        assert "compile_time" in ds.data_vars
        assert "compute_time" in ds.data_vars
        assert "total_time" in ds.data_vars
        assert "pareto_k" in ds.data_vars
        assert "lbfgs_status_counts" in ds.data_vars
        assert "path_status_counts" in ds.data_vars

        # Check per-path data (paths/ prefix)
        assert "paths/lbfgs_niter" in ds.data_vars
        assert "paths/elbo_argmax" in ds.data_vars
        assert "paths/logP_mean" in ds.data_vars
        assert "paths/logQ_mean" in ds.data_vars
        assert "paths/final_sample" in ds.data_vars

        # Verify path dimension
        assert "path" in ds.dims
        assert ds.sizes["path"] == 3
        assert ds["paths/lbfgs_niter"].shape == (3,)

        # Check config data (config/ prefix)
        assert "config/num_draws" in ds.data_vars
        assert "config/maxcor" in ds.data_vars
        assert "config/maxiter" in ds.data_vars
        assert ds["config/num_draws"].values == 100
        assert ds["config/maxcor"].values == 5

        # Check no diagnostics data when store_diagnostics=False
        diagnostics_vars = [k for k in ds.data_vars.keys() if k.startswith("diagnostics/")]
        assert len(diagnostics_vars) == 0

        # Test with diagnostics
        ds_with_diag = multipathfinder_result_to_xarray(result, model=None, store_diagnostics=True)

        # Check diagnostics data (diagnostics/ prefix)
        assert "diagnostics/logP_full" in ds_with_diag.data_vars
        assert "diagnostics/logQ_full" in ds_with_diag.data_vars
        assert "diagnostics/samples_full" in ds_with_diag.data_vars

        # Verify diagnostics shapes
        assert ds_with_diag["diagnostics/logP_full"].shape == (3, 100)
        assert ds_with_diag["diagnostics/samples_full"].shape == (3, 100, 2)


class TestAddPathfinderToInferenceData:
    """Tests for adding pathfinder results to InferenceData."""

    def test_add_to_inference_data(self):
        """Test adding pathfinder results to InferenceData object."""
        pytest.importorskip("arviz")

        import arviz as az

        from pymc_extras.inference.pathfinder.idata import add_pathfinder_to_inference_data

        # Create mock InferenceData
        posterior = xr.Dataset({"x": (["chain", "draw"], np.random.normal(0, 1, (1, 100)))})
        idata = az.InferenceData(posterior=posterior)

        # Create mock result with proper single-path status values
        # (Note: MockMultiPathfinderResult isn't a real MultiPathfinderResult,
        #  so it will be treated as single-path by add_pathfinder_to_inference_data)
        result = MockMultiPathfinderResult(
            samples=np.random.normal(0, 1, (2, 50, 1)),
            num_paths=2,
            num_draws=100,
            lbfgs_status=LBFGSStatus.CONVERGED,  # Single enum value, not Counter
            path_status=PathStatus.SUCCESS,  # Single enum value, not Counter
        )

        # Add pathfinder groups
        idata_updated = add_pathfinder_to_inference_data(idata, result, model=None)

        # Check groups were added
        # Note: Since MockMultiPathfinderResult is not a real MultiPathfinderResult,
        # it gets treated as a single-path result, so only 'pathfinder' group is added
        groups = list(idata_updated.groups())
        assert "posterior" in groups
        assert "pathfinder" in groups
        # pathfinder_paths is only created for true MultiPathfinderResult instances


class TestDiagnosticsAndConfigGroups:
    """Tests for diagnostics and config nested within consolidated pathfinder group."""

    def test_config_data_integration(self):
        """Test that config data is integrated into consolidated pathfinder group."""
        pytest.importorskip("arviz")

        import arviz as az

        from pymc_extras.inference.pathfinder.idata import add_pathfinder_to_inference_data
        from pymc_extras.inference.pathfinder.pathfinder import PathfinderConfig

        # Create mock InferenceData
        posterior = xr.Dataset({"x": (["chain", "draw"], np.random.normal(0, 1, (1, 100)))})
        idata = az.InferenceData(posterior=posterior)

        # Create mock config
        config = PathfinderConfig(
            num_draws=1000,
            maxcor=5,
            maxiter=100,
            ftol=1e-5,
            gtol=1e-8,
            maxls=1000,
            jitter=2.0,
            epsilon=1e-8,
            num_elbo_draws=10,
        )

        # Test with MultiPathfinderResult that has config
        result = MockMultiPathfinderResult(
            samples=np.random.normal(0, 1, (2, 50, 1)),
            num_paths=2,
            pathfinder_config=config,
            lbfgs_status=Counter({LBFGSStatus.CONVERGED: 2}),
            path_status=Counter({PathStatus.SUCCESS: 2}),
        )

        # Add pathfinder group
        idata_updated = add_pathfinder_to_inference_data(idata, result, model=None)

        # Check that we only have one pathfinder group
        groups = list(idata_updated.groups())
        assert "pathfinder" in groups
        assert "pathfinder_config" not in groups  # No separate config group

        # Check config data is nested within pathfinder group with config/ prefix
        assert "config/num_draws" in idata_updated.pathfinder.data_vars
        assert "config/maxcor" in idata_updated.pathfinder.data_vars
        assert "config/maxiter" in idata_updated.pathfinder.data_vars
        assert idata_updated.pathfinder["config/num_draws"].values == 1000
        assert idata_updated.pathfinder["config/maxcor"].values == 5

    def test_diagnostics_data_integration(self):
        """Test that diagnostics data is integrated into consolidated pathfinder group."""
        pytest.importorskip("arviz")

        import arviz as az

        from pymc_extras.inference.pathfinder.idata import add_pathfinder_to_inference_data

        # Create mock InferenceData
        posterior = xr.Dataset({"x": (["chain", "draw"], np.random.normal(0, 1, (1, 100)))})
        idata = az.InferenceData(posterior=posterior)

        # Create mock result with diagnostic data
        result = MockMultiPathfinderResult(
            samples=np.random.normal(0, 1, (2, 50, 3)),  # 2 paths, 50 draws, 3 params
            logP=np.random.normal(-10, 1, (2, 50)),  # Per-path, per-draw logP
            logQ=np.random.normal(-11, 1, (2, 50)),  # Per-path, per-draw logQ
            lbfgs_niter=np.array([30, 40]),
            elbo_argmax=np.array([15, 25]),
            lbfgs_status=Counter({LBFGSStatus.CONVERGED: 2}),
            path_status=Counter({PathStatus.SUCCESS: 2}),
            num_paths=2,
        )

        # Test with add_pathfinder_to_inference_data and store_diagnostics=True
        idata_updated = add_pathfinder_to_inference_data(
            idata, result, model=None, store_diagnostics=True
        )

        # Check that we only have one pathfinder group
        groups = list(idata_updated.groups())
        assert "pathfinder" in groups
        assert "pathfinder_diagnostics" not in groups  # No separate diagnostics group

        # Check diagnostics data is nested within pathfinder group with diagnostics/ prefix
        assert "diagnostics/logP_full" in idata_updated.pathfinder.data_vars
        assert "diagnostics/logQ_full" in idata_updated.pathfinder.data_vars
        assert "diagnostics/samples_full" in idata_updated.pathfinder.data_vars

        # Verify shapes
        assert idata_updated.pathfinder["diagnostics/logP_full"].shape == (2, 50)
        assert idata_updated.pathfinder["diagnostics/samples_full"].shape == (2, 50, 3)

    def test_no_diagnostics_when_store_false(self):
        """Test that diagnostics group is NOT created when store_diagnostics=False."""
        pytest.importorskip("arviz")

        import arviz as az

        from pymc_extras.inference.pathfinder.idata import add_pathfinder_to_inference_data

        # Create mock InferenceData
        posterior = xr.Dataset({"x": (["chain", "draw"], np.random.normal(0, 1, (1, 100)))})
        idata = az.InferenceData(posterior=posterior)

        # Create mock result with diagnostic data
        result = MockMultiPathfinderResult(
            samples=np.random.normal(0, 1, (2, 50, 3)),
            logP=np.random.normal(-10, 1, (2, 50)),
            logQ=np.random.normal(-11, 1, (2, 50)),
            num_paths=2,
        )

        # Test with store_diagnostics=False (default)
        idata_updated = add_pathfinder_to_inference_data(
            idata, result, model=None, store_diagnostics=False
        )

        # Check diagnostics group was NOT added
        groups = list(idata_updated.groups())
        assert "pathfinder_diagnostics" not in groups


def test_import_structure():
    """Test that all expected imports work."""
    # This test should pass even without full dependencies
    from pymc_extras.inference.pathfinder.idata import (
        _add_config_data,
        _add_diagnostics_data,
        _add_paths_data,
        _add_summary_data,
        add_pathfinder_to_inference_data,
        get_param_coords,
        multipathfinder_result_to_xarray,
        pathfinder_result_to_xarray,
    )

    # Check functions are callable
    assert callable(get_param_coords)
    assert callable(pathfinder_result_to_xarray)
    assert callable(multipathfinder_result_to_xarray)
    assert callable(add_pathfinder_to_inference_data)
    assert callable(_add_summary_data)
    assert callable(_add_paths_data)
    assert callable(_add_config_data)
    assert callable(_add_diagnostics_data)


if __name__ == "__main__":
    # Run basic import test
    test_import_structure()
    print("✓ Import structure test passed")



from pymc_extras.statespace.models import structural
from pymc_extras.statespace.models.DFM import BayesianDynamicFactor
from pymc_extras.statespace.models.ETS import BayesianETS
from pymc_extras.statespace.models.SARIMAX import BayesianSARIMAX
from pymc_extras.statespace.models.VARMAX import BayesianVARMAX

__all__ = [
    "BayesianSARIMAX",
    "BayesianVARMAX",
    "BayesianETS",
    "BayesianDynamicFactor",
    "structural",
]



import warnings

import numpy as np

from pytensor import tensor as pt

from pymc_extras.statespace.core.properties import (
    Coord,
    Data,
    Parameter,
    Shock,
    State,
)
from pymc_extras.statespace.models.structural.core import Component
from pymc_extras.statespace.models.utilities import validate_names
from pymc_extras.statespace.utils.constants import TIME_DIM


class Regression(Component):
    r"""
    Regression component for exogenous variables in a structural time series model

    Parameters
    ----------
    name : str | None, default "regression"
        A name for this regression component. Used to label dimensions and coordinates.

    state_names : list[str] | None, default None
        List of strings for regression coefficient labels. If provided, must be of length
        k_exog. If None and k_exog is provided, coefficients will be named
        "{name}_1, {name}_2, ...".

    observed_state_names : list[str] | None, default None
        List of strings for observed state labels. If None, defaults to ["data"].

    innovations : bool, default False
        Whether to include stochastic innovations in the regression coefficients,
        allowing them to vary over time. If True, coefficients follow a random walk.

    share_states: bool, default False
        Whether latent states are shared across the observed states. If True, there will be only one set of latent
        states, which are observed by all observed states. If False, each observed state has its own set of
        latent states.

    Notes
    -----
    This component implements regression with exogenous variables in a structural time series
    model. The regression component can be expressed as:

    .. math::
        y_t = \beta_t^T x_t + \epsilon_t

    Where :math:`y_t` is the dependent variable, :math:`x_t` is the vector of exogenous
    variables, :math:`\beta_t` is the vector of regression coefficients, and :math:`\epsilon_t`
    is the error term.

    When ``innovations=False`` (default), the coefficients are constant over time:
    :math:`\beta_t = \beta_0` for all t.

    When ``innovations=True``, the coefficients follow a random walk:
    :math:`\beta_{t+1} = \beta_t + \eta_t`, where :math:`\eta_t \sim N(0, \Sigma_\beta)`.

    The component supports both univariate and multivariate regression. In the multivariate
    case, separate coefficients are estimated for each endogenous variable (i.e time series).

    Examples
    --------
    Simple regression with constant coefficients:

    .. code:: python

        from pymc_extras.statespace import structural as st
        import pymc as pm
        import pytensor.tensor as pt

        trend = st.LevelTrend(order=1, innovations_order=1)
        regression = st.Regression(k_exog=2, state_names=['intercept', 'slope'])
        ss_mod = (trend + regression).build()

        with pm.Model(coords=ss_mod.coords) as model:
            # Prior for regression coefficients
            betas = pm.Normal('betas', dims=ss_mod.param_dims['beta_regression'])

            # Prior for trend innovations
            sigma_trend = pm.Exponential('sigma_trend', 1)

            ss_mod.build_statespace_graph(data)
            idata = pm.sample()

    Multivariate regression with time-varying coefficients:
    - There are 2 exogenous variables (price and income effects)
    - There are 2 endogenous variables (sales and revenue)
    - The regression coefficients are allowed to vary over time (`innovations=True`)

    .. code:: python

        regression = st.Regression(
            k_exog=2,
            state_names=['price_effect', 'income_effect'],
            observed_state_names=['sales', 'revenue'],
            innovations=True
        )

        with pm.Model(coords=ss_mod.coords) as model:
            betas = pm.Normal('betas', dims=ss_mod.param_dims['beta_regression'])

            # Innovation variance for time-varying coefficients
            sigma_beta = pm.Exponential('sigma_beta', 1, dims=ss_mod.param_dims['sigma_beta_regression'])

            ss_mod.build_statespace_graph(data)
            idata = pm.sample()
    """

    def __init__(
        self,
        name: str | None = "regression",
        state_names: list[str] | None = None,
        observed_state_names: list[str] | None = None,
        innovations=False,
        share_states: bool = False,
    ):
        self.share_states = share_states

        if observed_state_names is None:
            observed_state_names = ["data"]

        self.innovations = innovations
        validate_names(state_names, var_name="state_names", optional=False)
        k_exog = len(state_names)

        k_states = k_exog
        k_endog = len(observed_state_names)
        k_posdef = k_exog

        super().__init__(
            name=name,
            k_endog=k_endog,
            k_states=k_states * k_endog if not share_states else k_states,
            k_posdef=k_posdef * k_endog if not share_states else k_posdef,
            base_state_names=state_names,
            share_states=share_states,
            base_observed_state_names=observed_state_names,
            measurement_error=False,
            combine_hidden_states=False,
            obs_state_idxs=np.ones(k_states),
        )

    def set_states(self) -> State | tuple[State, ...] | None:
        base_names = self.base_state_names
        observed_state_names = self.base_observed_state_names

        if self.share_states:
            state_names = [f"{name}[{self.name}_shared]" for name in base_names]
        else:
            state_names = [
                f"{name}[{obs_name}]" for obs_name in observed_state_names for name in base_names
            ]

        hidden_states = [State(name=name, observed=False, shared=True) for name in state_names]
        observed_states = [
            State(name=name, observed=True, shared=False) for name in observed_state_names
        ]
        return *hidden_states, *observed_states

    def set_parameters(self) -> Parameter | tuple[Parameter, ...] | None:
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog
        k_states = self.k_states // k_endog_effective

        beta_parameter = Parameter(
            name=f"beta_{self.name}",
            shape=(k_endog_effective, k_states) if k_endog_effective > 1 else (k_states,),
            dims=(
                (f"endog_{self.name}", f"state_{self.name}")
                if k_endog_effective > 1
                else (f"state_{self.name}",)
            ),
            constraints=None,
        )

        params_container = [beta_parameter]

        if self.innovations:
            sigma_parameter = Parameter(
                name=f"sigma_beta_{self.name}",
                shape=(k_states,),
                dims=(f"state_{self.name}",),
                constraints="Positive",
            )

            params_container.append(sigma_parameter)

        return tuple(params_container)

    def set_data_info(self) -> Data | tuple[Data, ...] | None:
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog
        k_states = self.k_states // k_endog_effective

        data_prop = Data(
            name=f"data_{self.name}",
            shape=(None, k_states),
            dims=(TIME_DIM, f"state_{self.name}"),
            is_exogenous=True,
        )
        return (data_prop,)

    def set_shocks(self) -> Shock | tuple[Shock, ...] | None:
        base_names = self.base_state_names

        if self.share_states:
            shock_names = [f"{state_name}_shared" for state_name in base_names]
        else:
            shock_names = base_names

        return tuple(Shock(name=name) for name in shock_names)

    def set_coords(self) -> tuple[Coord, ...] | None:
        regression_state_coord = Coord(
            dimension=f"state_{self.name}", labels=tuple(self.base_state_names)
        )
        endogenous_state_coord = Coord(
            dimension=f"endog_{self.name}", labels=self.observed_state_names
        )

        return regression_state_coord, endogenous_state_coord

    def make_symbolic_graph(self) -> None:
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog

        k_states = self.k_states // k_endog_effective

        betas = self.make_and_register_variable(
            f"beta_{self.name}", shape=(k_endog, k_states) if k_endog_effective > 1 else (k_states,)
        )
        regression_data = self.make_and_register_data(f"data_{self.name}", shape=(None, k_states))

        self.ssm["initial_state", :] = betas.ravel()
        self.ssm["transition", :, :] = pt.eye(self.k_states)
        self.ssm["selection", :, :] = pt.eye(self.k_states)

        if self.share_states:
            self.ssm["design"] = pt.specify_shape(
                pt.join(1, *[pt.expand_dims(regression_data, 1) for _ in range(k_endog)]),
                (None, k_endog, self.k_states),
            )
        else:
            Z = pt.linalg.block_diag(*[pt.expand_dims(regression_data, 1) for _ in range(k_endog)])
            self.ssm["design"] = pt.specify_shape(
                Z, (None, k_endog, regression_data.type.shape[1] * k_endog)
            )

        if self.innovations:
            sigma_beta = self.make_and_register_variable(
                f"sigma_beta_{self.name}",
                (k_states,) if k_endog_effective == 1 else (k_endog, k_states),
            )
            row_idx, col_idx = np.diag_indices(self.k_states)
            self.ssm["state_cov", row_idx, col_idx] = sigma_beta.ravel() ** 2


def __getattr__(name: str):
    if name == "RegressionComponent":
        warnings.warn(
            "RegressionComponent is deprecated and will be removed in a future release. "
            "Use Regression instead.",
            FutureWarning,
            stacklevel=2,
        )
        return Regression
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
